# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import ast
import sys
from collections import defaultdict
from .context import Context
from .utils import diss
from ..cross_version import cross_compile, StringIO
from ..persistence import persistence


class FunctionVisitor(ast.NodeVisitor):
    'Identifies the function declarations and related data'
    code = None
    functions = {}

    # Temporary attributes for recursive data collection
    contexts = [Context('(global)')]
    names = None
    lineno = None

    def __init__(self, metascript):
        self.code = metascript['code'].decode('utf-8').split('\n')
        self.metascript = metascript
        self.result = None

    @property
    def namespace(self):
        return '.'.join(context.name for context in self.contexts[1:])

    def generic_visit(self, node):
        # Delegation, but collecting the current line number
        try:
            self.lineno = node.lineno
        except:
            pass
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ClassDef(self, node): # ignoring classes
        self.contexts.append(Context(node.name))
        self.generic_visit(node)
        self.contexts.pop()

    def visit_FunctionDef(self, node):
        self.contexts.append(Context(node.name))
        self.generic_visit(node)
        code_hash = persistence.put(
            '\n'.join(self.code[node.lineno - 1:self.lineno]).encode('utf-8'))
        self.functions[self.namespace] = self.contexts[-1].to_tuple(code_hash)
        self.contexts.pop()

    def visit_arguments(self, node):
        self.names = []
        self.generic_visit(node)
        self.contexts[-1].arguments.extend(self.names)

    def visit_Global(self, node):
        self.contexts[-1].global_vars.extend(node.names)
        self.generic_visit(node)

    def call(self, node):
        func = node.func
        if isinstance(func, ast.Name): # collecting only direct function call
            self.contexts[-1].calls.append(func.id)

    def visit_Call(self, node):
        self.call(node)
        self.generic_visit(node)

    def visit_Name(self, node):
        if self.names != None:
            self.names.append(node.id)
        self.generic_visit(node)

    def teardown(self):
        pass

    def extract_disasm(self):
        self.metascript['compiled'] = cross_compile(
            self.metascript['code'], self.metascript['path'], 'exec')
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        diss(self.metascript['compiled'], recurse=True)
        #dis.dis(self.metascript['compiled'])
        #for fn in self.metascript['compiled'].co_consts:
        #    dis.dis(fn)
        sys.stdout = old_stdout

        self.disasm = mystdout.getvalue().split('\n')

        # Sort lines
        lines = defaultdict(list)
        line = -1
        for disasm in self.disasm:
            num = disasm[:8].strip()
            if num:
                line = int(num)
            lines[line].append(disasm)

        keys = sorted(lines.keys())
        self.disasm = []
        for line in keys:
            self.disasm.extend(lines[line])


# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import argparse

from .. import utils
from ..persistence import persistence
from ..models.trial import Trial
from ..models.trial_prolog import TrialProlog
from .command import Command


def int_or_type(string):
    try:
        return int(string)
    except ValueError:
        if "diff" in string:
            splitted = string.split(':')
            if len(splitted) < 3:
                raise argparse.ArgumentTypeError("you must diff two trials")
            return splitted
        return string


class Export(Command):

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('-r', '--rules', action='store_true',
                help='also exports inference rules')
        add_arg('-i', '--ipython', action='store_true',
                help='export ipython notebook file')
        add_arg('trial', type=int_or_type, nargs='?',
                help='trial id or none for last trial. If you are generation '
                     'ipython notebook files, it is also possible to use "history"'
                     'or "diff:<trial_id_1>:<trial_id_2>"')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        if not args.ipython:
            trial = Trial(trial_id=args.trial, exit=True)
            trial_prolog = TrialProlog(trial)
            print(trial_prolog.export_text_facts())
            if args.rules:
                print(trial_prolog.export_rules())
        else:
            import IPython.nbformat.current as nbf
            if args.trial == "history":
                nb = nbf.reads((u"%load_ext noworkflow\n"
                                u"import noworkflow.now.ipython as nip\n"
                                u"# <codecell>\n"
                                u"history = nip.History()\n"
                                u"# history.graph_width = 700\n"
                                u"# history.graph_height = 300\n"
                                u"# history.script = '*'\n"
                                u"# history.execution = '*'\n"
                                u"# <codecell>\n"
                                u"history").format(args.trial), 'py')
                with open('History.ipynb'.format(args.trial),'w') as ipynb:
                    nbf.write(nb, ipynb, 'ipynb')
            elif isinstance(args.trial, list):
                nb = nbf.reads((u"%load_ext noworkflow\n"
                                u"import noworkflow.now.ipython as nip\n"
                                u"# <codecell>\n"
                                u"diff = nip.Diff({1}, {2})\n"
                                u"# diff.graph_type = 0\n"
                                u"# diff.display_mode = 0\n"
                                u"# diff.graph_width = 500\n"
                                u"# diff.graph_height = 500\n"
                                u"# <codecell>\n"
                                u"diff").format(*args.trial), 'py')
                with open('Diff-{1}-{2}.ipynb'.format(*args.trial),'w') as ipynb:
                    nbf.write(nb, ipynb, 'ipynb')
            else:
                nb = nbf.reads((u"%load_ext noworkflow\n"
                                u"import noworkflow.now.ipython as nip\n"
                                u"# <codecell>\n"
                                u"trial = nip.Trial({})\n"
                                u"# trial.graph_type = 0\n"
                                u"# trial.graph_width = 500\n"
                                u"# trial.graph_height = 500\n"
                                u"# <codecell>\n"
                                u"trial").format(args.trial), 'py')
                with open('Trial-{}.ipynb'.format(args.trial),'w') as ipynb:
                    nbf.write(nb, ipynb, 'ipynb')



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import json

from .provider import Provider, row_to_dict


class RestoreProvider(Provider):

    def remove_head(self, trial_id):
        with self.db_conn as db:
            db.execute('''DELETE FROM head
                          WHERE trial_id=?''', (trial_id,))

    def load_head(self, script):
        try:
            with self.db_conn as db:
                (an_id,) = db.execute(
                    '''SELECT trial_id
                       FROM head
                       WHERE script=?''', (script,)).fetchone()
                return an_id
        except TypeError:
            return None

    def load_parent_id(self, script, remove=True, parent_required=False):
        an_id = self.load_head(script)
        if an_id and remove:
            self.remove_head(an_id)
        elif not an_id:
            an_id = self.last_trial_id(script=script,
                                       parent_required=parent_required)
        return an_id

    def store_parent(self, script, trial_id):
        an_id = self.load_head(script)
        with self.db_conn as db:
            if an_id:
                db.execute('''UPDATE head
                              SET trial_id=?
                              WHERE script=?''', (trial_id, script))
            else:
                db.execute("""INSERT INTO head(trial_id, script)
                              VALUES(?,?)""", [trial_id, script])



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import sys
import traceback
import argparse

from .. import prov_definition
from .. import prov_deployment
from .. import prov_execution
from .. import utils
from ..persistence import persistence
from ..cross_version import cross_compile
from .command import Command

LAST_TRIAL = '.last_trial'


def non_negative(string):
    value = int(string)
    if value < 0:
        raise argparse.ArgumentTypeError(
            "%s is not a non-negative integer value" % string)
    return value


class Run(Command):

    def add_arguments(self):
        add_arg = self.add_argument
        add_cmd = self.add_argument_cmd
        add_arg('-v', '--verbose', action='store_true',
                help='increase output verbosity')
        add_arg('-b', '--bypass-modules', action='store_true',
                help='bypass module dependencies analysis, assuming that no '
                     'module changes occurred since last execution')
        add_arg('-c', '--depth-context', choices=['non-user', 'all'],
                default='non-user',
                help='functions subject to depth computation when capturing '
                     'activations (defaults to non-user)')
        add_arg('-d', '--depth', type=non_negative, default=1,
                help='depth for capturing function activations (defaults to '
                     '1)')
        add_arg('-e', '--execution-provenance', default="Profiler",
                choices=['Profiler', 'InspectProfiler', 'Tracer'],
                help='execution provenance provider. (defaults to Profiler)')
        add_arg('--disasm', action='store_true',
                help='show script disassembly')
        add_arg('--name', type=str,
                help="set branch name used for tracking history")
        add_arg('--dir', type=str,
                help='set project path. The noworkflow database folder will '
                     'be created in this path. Default to script directory')
        add_cmd('--create_last', action='store_true')
        add_cmd('script', nargs=argparse.REMAINDER,
                help='Python script to be executed')

    def execute(self, args):
        utils.verbose = args.verbose
        utils.print_msg('removing noWorkflow boilerplate')

        args_script = args.script
        args.script = os.path.realpath(args.script[0])

        if not os.path.exists(args.script):  # TODO: check this using argparse
            utils.print_msg('the script does not exist', True)
            sys.exit(1)

        script_dir = args.dir or os.path.dirname(args.script)

        # Replace now's dir with script's dir in front of module search path.
        sys.path[0] = os.path.dirname(args.script)

        # Clear argv
        sys.argv = args_script
        # Clear up the __main__ namespace
        import __main__
        __main__.__dict__.clear()
        __main__.__dict__.update({'__name__'    : '__main__',
                                  '__file__'    : args.script,
                                  '__builtins__': __builtins__,
                                 })

        with open(args.script, 'rb') as f:
            metascript = {
                'trial_id': None,
                'code': f.read(),
                'path': args.script,
                'compiled': None,
                'definition': None,
                'name': args.name or os.path.basename(sys.argv[0])
            }
        try:
            self.run(script_dir, args, metascript, __main__)
        finally:
            if args.create_last:
                tmp = os.path.join(os.path.dirname(args.script), LAST_TRIAL)
                with open(tmp, 'w') as f:
                    f.write(str(metascript['trial_id']))

    def run(self, script_dir, args, metascript, __main__):

        utils.print_msg('setting up local provenance store')
        persistence.connect(script_dir)

        utils.print_msg('collecting definition provenance')
        prov_definition.collect_provenance(args, metascript)

        utils.print_msg('collecting deployment provenance')
        prov_deployment.collect_provenance(args, metascript)

        utils.print_msg('collection execution provenance')
        prov_execution.enable(args, metascript)

        utils.print_msg('  executing the script')
        try:
            if metascript['compiled'] is None:
                metascript['compiled'] = cross_compile(
                    metascript['code'], metascript['path'], 'exec')
            exec(metascript['compiled'], __main__.__dict__)

        except SystemExit as ex:
            prov_execution.disable()
            utils.print_msg(
                'the execution exited via sys.exit(). Exit status: {}'
                ''.format(ex.code), ex.code > 0)
        except Exception as e:
            prov_execution.disable()
            print(e)
            utils.print_msg(
                'the execution finished with an uncaught exception. {}'
                ''.format(traceback.format_exc()), True)
        else:
            # TODO: exceptions should be registered as return from the
            # activation and stored in the database. We are currently ignoring
            # all the activation tree when exceptions are raised.
            prov_execution.disable()
            prov_execution.store()
            utils.print_msg(
                'the execution of trial {} finished successfully'
                ''.format(metascript['trial_id']))

        return prov_execution.provider



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import webbrowser
import threading

from ..persistence import persistence
from .command import Command

class Vis(Command):

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('-p', '--port', nargs='?', type=int, default=5000,
                help='sets server port')
        add_arg('-d', '--debug', action='store_true',
                help='debug mode')
        add_arg('-b', '--browser', action='store_true',
                help='opens browser')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        if args.browser:
            url = "http://127.0.0.1:{0}".format(args.port)
            print(url)
            threading.Timer(1.25, lambda: webbrowser.open(url)).start()
        from ..vis.views import app
        app.dir = args.dir or os.getcwd()
        app.run(port=args.port, debug=args.debug)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from IPython.utils.text import DollarFormatter

import argparse

from ..models import TrialProlog, Trial
from .command import IpythonCommandMagic


class NowProlog(IpythonCommandMagic):
    """Queries the provenance database with Prolog

    Examples
    --------
    ::
        In  [1]: %%now_prolog 1
            ...: duration(1, z, X)
        Out [1]: [{'X': 0.10173702239990234},
            ...:  {'X': 0.10082292556762695},
            ...:  {'X': 0.1021270751953125},
            ...:  {'X': 0.10217714309692383}]

        In  [2]: %%now_prolog --result tupleit
            ...: SELECT * FROM trial

        In  [3]: duration(1, z, X)
        Out [3]: <generator object __call__ at 0x7f79ed329f50>
    """

    def add_arguments(self):
        super(NowProlog, self).add_arguments()
        add_arg = self.add_argument
        add_arg('--result', type=str,
                help="""The variable in which the result will be stored""")
        add_arg('trials', nargs=argparse.REMAINDER,
                help='export trial facts')

    def execute(self, func, line, cell, magic_cls):
        f = DollarFormatter()
        cell = f.vformat(cell, args=[], kwargs=magic_cls.shell.user_ns.copy())
        _, args = self.arguments(func, line)
        for trial in args.trials:
            Trial(int(trial)).trial_prolog.load_cli_facts()
        result = TrialProlog.prolog_query(cell)
        if args.result:
            magic_cls.shell.user_ns[args.result] = result
        else:
            return list(result)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
import pyposast
from ...now.cross_version import bytes_string
from ...now.persistence import persistence
from ...now.prov_definition.slicing_visitor import SlicingVisitor
from ...now.prov_definition.utils import FunctionCall


persistence.put = lambda x: None
NAME = '<unknown>'


class TestSlicingDependencies(unittest.TestCase):

        def dependencies(self, line):
            return self.visitor.dependencies[line]

        def call(self, line, col):
            return self.visitor.function_calls[line][col]

        def parse(self, code):
            metascript = {
                'name': 'name',
                'code': bytes_string(code, 'utf-8'),
                'path': NAME,
                'compiled': None,
            }
            self.visitor = SlicingVisitor(metascript)
            self.visitor.metascript['code'] = code
            return pyposast.parse(code)

        def test_simple_assignment(self):
            tree = self.parse("a = 1\n"
                             "b = a")
            self.visitor.visit(tree)
            self.assertIn('a', self.dependencies(1))
            self.assertEqual(['a'], self.dependencies(2)['b'])

        def test_unpack_assignment(self):
            tree = self.parse("a, b = (a + c), d")
            self.visitor.visit(tree)
            self.assertEqual(['a', 'c'], self.dependencies(1)['a'])
            self.assertEqual(['d'], self.dependencies(1)['b'])

        def test_multiple_assignment(self):
            tree = self.parse("c = a, b = d, e")
            self.visitor.visit(tree)
            self.assertEqual(['d'], self.dependencies(1)['a'])
            self.assertEqual(['e'], self.dependencies(1)['b'])
            self.assertEqual(['d', 'e'], self.dependencies(1)['c'])

        def test_for(self):
            tree = self.parse("for i in a + [b, c]:\n"
                             "    pass")
            self.visitor.visit(tree)
            self.assertEqual(['a', 'b', 'c'], self.dependencies(1)['i'])

        def test_for2(self):
            tree = self.parse("for i, j in [b, c]:\n"
                             "    pass")
            self.visitor.visit(tree)
            self.assertEqual(['b', 'c'], self.dependencies(1)['i'])
            self.assertEqual(['b', 'c'], self.dependencies(1)['j'])


        def test_augmented_assignment(self):
            tree = self.parse("a += 1\n"
                             "b += a")
            self.visitor.visit(tree)
            self.assertEqual(['a'], self.dependencies(1)['a'])
            self.assertEqual(['a', 'b'], self.dependencies(2)['b'])

        def test_attribute_assignment(self):
            tree = self.parse("a.attr.attr2 = b")
            self.visitor.visit(tree)
            self.assertEqual(['b'], self.dependencies(1)['a'])
            self.assertEqual(['b'], self.dependencies(1)['a.attr'])
            self.assertEqual(['b'], self.dependencies(1)['a.attr.attr2'])

        def test_subscript_assignment(self):
            tree = self.parse("a[c] = b")
            self.visitor.visit(tree)
            self.assertEqual(['b'], self.dependencies(1)['a'])
            self.assertFalse(self.dependencies(1)['c'])

        def test_complex_subscript_assignment(self):
            tree = self.parse("a.attr.attr5[c.attr2[d.attr3]].attr4 = b")
            self.visitor.visit(tree)
            self.assertEqual(['b'], self.dependencies(1)['a'])
            self.assertEqual(['b'], self.dependencies(1)['a.attr'])
            self.assertEqual(['b'], self.dependencies(1)['a.attr.attr5'])
            self.assertFalse(self.dependencies(1)['a.attr.attr5.attr4'])
            self.assertFalse(self.dependencies(1)['c'])
            self.assertFalse(self.dependencies(1)['c.attr2'])
            self.assertFalse(self.dependencies(1)['d'])
            self.assertFalse(self.dependencies(1)['d.attr3'])

        def test_lambda_assignment(self):
            tree = self.parse("a = (lambda x: x + b)(c)")
            self.visitor.visit(tree)
            self.assertEqual([(1, 24)], self.dependencies(1)['a'])
            self.assertEqual([['c']], self.call(1, 24).args)

        def test_lambda2_assignment(self):
            tree = self.parse("a = (lambda x: (lambda y: x + y))(b)(c)")
            self.visitor.visit(tree)
            self.assertEqual([(1, 39)], self.dependencies(1)['a'])
            self.assertEqual([['c']], self.call(1, 39).args)
            self.assertEqual([(1, 36)], self.call(1, 39).func)
            self.assertEqual([['b']], self.call(1, 36).args)

        def test_list_comprehension_assignment(self):
            tree = self.parse("a = [i + b for i in c if i + d == b]")
            self.visitor.visit(tree)
            self.assertEqual(['c', 'd', 'b', 'b'], self.dependencies(1)['a'])

        def test_list_comprehension2_assignment(self):
            tree = self.parse("a = [i + j for i in c for j in d]")
            self.visitor.visit(tree)
            self.assertEqual(['c', 'd'], self.dependencies(1)['a'])

        def test_set_comprehension_assignment(self):
            tree = self.parse("a = {i for i in c}")
            self.visitor.visit(tree)
            self.assertEqual(['c'], self.dependencies(1)['a'])

        def test_generator_assignment(self):
            tree = self.parse("a = sum(i for i in c)")
            self.visitor.visit(tree)
            self.assertEqual((1, 21), self.dependencies(1)['a'][0])

        def test_dict_comprehension_assignment(self):
            tree = self.parse("a = {i:i**b for i in c}")
            self.visitor.visit(tree)
            self.assertEqual(['c', 'b'], self.dependencies(1)['a'])

        def test_function_call_assignment(self):
            tree = self.parse("a = fn(b=c)")
            self.visitor.visit(tree)
            self.assertEqual((1, 11), self.dependencies(1)['a'][0])
            self.assertEqual({'b':['c']}, self.call(1, 11).keywords)

        def test_nested_call(self):
            tree = self.parse("a = fn(fn(x))")
            self.visitor.visit(tree)
            self.assertEqual((1, 13), self.dependencies(1)['a'][0])
            self.assertEqual([[(1, 12)]], self.call(1, 13).args)

        def test_while(self):
            tree = self.parse("while i:\n"
                             "    b = c\n"
                             "else:\n"
                             "    c = d")
            self.visitor.visit(tree)
            self.assertEqual(['c', 'i'], self.dependencies(2)['b'])
            self.assertEqual(['d', 'i'], self.dependencies(4)['c'])

        def test_if(self):
            tree = self.parse("if i:\n"
                             "    b = c\n"
                             "else:\n"
                             "    c = d")
            self.visitor.visit(tree)
            self.assertEqual(['c', 'i'], self.dependencies(2)['b'])
            self.assertEqual(['d', 'i'], self.dependencies(4)['c'])

        def test_nested(self):
            tree = self.parse("if i:\n"
                             "    if j:\n"
                             "        a = x\n"
                             "    b = y\n"
                             "c = z")
            self.visitor.visit(tree)
            self.assertEqual({'x', 'i', 'j'}, set(self.dependencies(3)['a']))
            self.assertEqual(['y', 'i'], self.dependencies(4)['b'])
            self.assertEqual(['z'], self.dependencies(5)['c'])

        def test_for_independent(self):
            tree = self.parse("for i in a:\n"
                             "    b = c")
            self.visitor.visit(tree)
            self.assertEqual(['c'], self.dependencies(2)['b'])

        def test_for_dependent(self):
            tree = self.parse("for i in a:\n"
                             "    b = b + c")
            self.visitor.visit(tree)
            self.assertEqual(['b', 'c', 'i'], self.dependencies(2)['b'])

        def test_for_dependent_augment(self):
            tree = self.parse("for i in a:\n"
                             "    b += c")
            self.visitor.visit(tree)
            self.assertEqual(['c', 'b', 'i'], self.dependencies(2)['b'])

        def test_return(self):
            code = ("def f(a):\n"
                    "    return a")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(['a'], self.dependencies(2)['return'])

        def test_yield(self):
            code = ("def f(a):\n"
                    "    yield a")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(['a'], self.dependencies(2)['return'])

        def test_return2(self):
            code = ("def f(a):\n"
                    "    if a:\n"
                    "        return 2")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(['a'], self.dependencies(3)['return'])

        def test_return3(self):
            code = ("def f(a):\n"
                    "    if a:\n"
                    "        return")
            tree = self.parse(code)
            self.visitor.visit(tree)
            self.assertEqual(['a'], self.dependencies(3)['return'])



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import sys

from .. import utils
from ..persistence import persistence
from ..models.diff import Diff as DiffModel
from .command import Command


class Diff(Command):

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('trial', type=int, nargs=2,
                help='trial id to be compared')
        add_arg('-m', '--modules', action='store_true',
                help='compare module dependencies')
        add_arg('-e', '--environment', action='store_true',
                help='compare environment conditions')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        diff = DiffModel(args.trial[0], args.trial[1], exit=True)
        self.diff_trials(diff)

        if args.modules:
            self.diff_modules(diff)

        if args.environment:
            self.diff_environment(diff)

    def diff_trials(self, diff):
        utils.print_msg('trial diff:', True)
        for key, values in diff.trial().items():
            print('  {} changed from {} to {}'.format(
                key, values[0], values[1]))
        print()

    def diff_modules(self, diff):
        (added, removed, replaced) = diff.modules()

        utils.print_msg('{} modules added:'.format(len(added),), True)
        utils.print_modules(added)
        print()

        utils.print_msg('{} modules removed:'.format(len(removed),), True)
        utils.print_modules(removed)
        print()

        utils.print_msg('{} modules replaced:'.format(len(replaced),), True)
        for (module_removed, module_added) in replaced:
            print('  Name: {}'.format(module_removed['name'],))
            self.diff_dict(module_removed, module_added)
            print()

    def diff_environment(self, diff):
        (added, removed, replaced) = diff.environment()

        utils.print_msg('{} environment attributes added:'.format(
            len(added),), True)
        utils.print_environment_attrs(added)
        print()

        utils.print_msg('{} environment attributes removed:'.format(
            len(removed),), True)
        utils.print_environment_attrs(removed)
        print()

        utils.print_msg('{} environment attributes replaced:'.format(
            len(replaced),), True)
        for (attr_removed, attr_added) in replaced:
            print('  Environment attribute {} changed from {} to {}'.format(
                attr_removed['name'],
                attr_removed['value'],
                attr_added['value']))



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .provider import Provider

class RunProvider(Provider):

    def store_trial(self, start, script, code, arguments, bypass_modules,
                    run=True):
        code_hash = self.put(code)
        parent_id = self.load_parent_id(script, parent_required=True)
        inherited_id = None
        if bypass_modules:
            inherited_id = self.last_trial_id_without_inheritance()
        with self.db_conn as db:
            trial_id = db.execute(
                """INSERT INTO trial (start, script, code_hash, arguments,
                                      inherited_id, parent_id, run)
                   VALUES (?, ?, ?, ?, ?, ?, ?)
                """, (start, script, code_hash, arguments,
                      inherited_id, parent_id, run)).lastrowid
        return trial_id

    def update_trial(self, trial_id, finish, function_activation):
        self.store_function_activation(trial_id, function_activation, None)
        with self.db_conn as db:
            db.execute(
                """UPDATE trial
                   SET finish = ?
                   WHERE id = ?""", (finish, trial_id))

    def store_objects(self, objects, obj_type, function_def_id):
        with self.db_conn as db:
            db.executemany(
                """INSERT INTO object(name, type, function_def_id)
                   VALUES (?, ?, ?)""",
                ((name, obj_type, function_def_id) for name in objects)
            )

    def store_function_defs(self, trial_id, functions):
        with self.db_conn as db:
            for name, defs in functions.items():
                arguments, global_vars, calls, code_hash = defs
                function_def_id = db.execute(
                    """INSERT INTO function_def(name, code_hash, trial_id)
                       VALUES (?, ?, ?)
                    """,(name, code_hash, trial_id)).lastrowid
                self.store_objects(arguments, 'ARGUMENT', function_def_id)
                self.store_objects(global_vars, 'GLOBAL', function_def_id)
                self.store_objects(calls, 'FUNCTION_CALL', function_def_id)

    def extract_function_activation(self, trial_id, activation, caller_id,
                                    activation_id):
        return (
            activation_id, activation.name, activation.line,
            activation.return_value, activation.start, activation.finish,
            caller_id, trial_id,
        )

    def extract_object_values(self, object_values, obj_type, activation_id):
        for name in object_values:
            yield (name, object_values[name], obj_type, activation_id)

    def extract_file_accesses(self, trial_id, file_accesses, activation_id):
        for file_access in file_accesses:
            yield (
                file_access['name'],
                file_access['mode'],
                file_access['buffering'],
                file_access['content_hash_before'],
                file_access['content_hash_after'],
                file_access['timestamp'],
                activation_id,
                trial_id
            )

    def store_function_activation(self, trial_id, activation, caller_id):
        function_activations, object_values, file_accesses = [], [], []
        d = {
            'fid': self.function_activation_id_seq(),
            'activations': function_activations,
            'object_values': object_values,
            'file_accesses': file_accesses,
        }

        def add_activation(trial_id, act, caller_id):
            fid = d['fid']
            d['activations'].append(
                self.extract_function_activation(trial_id, act, caller_id, fid)
            )
            for object_value in self.extract_object_values(act.arguments,
                                                           'ARGUMENT', fid):
                d['object_values'].append(object_value)

            for object_value in self.extract_object_values(act.globals,
                                                           'GLOBAL', fid):
                d['object_values'].append(object_value)

            for file_access in self.extract_file_accesses(trial_id,
                                                          act.file_accesses,
                                                          fid):
                d['file_accesses'].append(file_access)

            d['fid'] += 1

            for inner_function_activation in act.function_activations:
                add_activation(trial_id, inner_function_activation, fid)

        add_activation(trial_id, activation, caller_id)


        with self.db_conn as db:
            db.executemany(
                """INSERT INTO function_activation(id, name, line, return,
                    start, finish, caller_id, trial_id)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
                function_activations
            )
            db.executemany(
                """INSERT INTO object_value(name, value, type,
                    function_activation_id)
                VALUES (?, ?, ?, ?)""",
                object_values
            )
            db.executemany(
                """INSERT INTO file_access(name, mode, buffering,
                    content_hash_before, content_hash_after, timestamp,
                    function_activation_id, trial_id)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
                file_accesses
            )

    def store_slicing(self, trial_id, variables, dependencies, usages):
        with self.db_conn as db:
            db.executemany(
                """INSERT INTO slicing_variable(trial_id, vid, name, line,
                    value, time)
                VALUES (?, ?, ?, ?, ?, ?)""",
                ((trial_id, v.id, v.name, v.line,
                    v.value, v.time) for v in variables)
            )
            db.executemany(
                """INSERT INTO slicing_dependency(trial_id, id, dependent,
                    supplier)
                VALUES (?, ?, ?, ?)""",
                ((trial_id, d.id, d.dependent,
                    d.supplier) for d in dependencies)
            )
            db.executemany(
                """INSERT INTO slicing_usage(trial_id, id, vid, name, line)
                VALUES (?, ?, ?, ?, ?)""",
                ((trial_id, u.id, u.vid, u.name, u.line) for u in usages)
            )



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Test now.formatter module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import unittest
from ..now.formatter import PrettyLines, Table


class PrettyMock(object):

    def __init__(self):
        self.result = ""

    def text(self, t):
        self.result += t


class TestFormatter(unittest.TestCase):
    """TestCase for now.formatter module"""

    def test_pretty_lines_repr_pretty_(self):
        lines = ["a", "b"]
        pmock = PrettyMock()
        plines = PrettyLines(lines)
        plines._repr_pretty_(pmock, None)
        self.assertEqual("a\nb", pmock.result)

    def test_pretty_lines_str_(self):
        lines = ["a", "b"]
        plines = PrettyLines(lines)
        self.assertEqual("a\nb", str(plines))

    def test_table_repr_html_with_header(self):
        table = [["a", "b"], ["1", "2"]]
        table = Table(table)
        self.assertEqual(
            "<table>"
            "<tr><th>a</th><th>b</th></tr>"
            "<tr><td>1</td><td>2</td></tr>"
            "</table><br>",
            table._repr_html_()
        )

    def test_table_repr_html_without_header(self):
        table = [["a", "b"], ["1", "2"]]
        table = Table(table)
        table.has_header = False
        self.assertEqual(
            "<table>"
            "<tr><td>a</td><td>b</td></tr>"
            "<tr><td>1</td><td>2</td></tr>"
            "</table><br>",
            table._repr_html_()
        )

    def test_table_repr_html_skip_header(self):
        table = [["a", "b"], ["1", "2"]]
        table = Table(table)
        table.show_header = False
        self.assertEqual(
            "<table>"
            "<tr><td>1</td><td>2</td></tr>"
            "</table><br>",
            table._repr_html_()
        )

    def test_table_str_(self):
        table = [["a", "bb"], ["1", "2"]]
        table = Table(table)
        self.assertEqual("a bb\n1  2\n", str(table))



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from ...cmd import Restore
from ...persistence import persistence

from .command import IpythonCommandMagic


class NowRestore(IpythonCommandMagic, Restore):
    """Restores trial and returns the script content

    Examples
    --------
    ::
        In [1]: content = %now_restore 2 --file
           ...: print(content)
        [now] File script2.py from trial 2 restored
        l = range(4)
        c = sum(l)
        print(c)

    """

    def add_arguments(self):
        super(NowRestore, self).add_arguments()
        add_arg = self.add_argument
        add_arg('--file', action="store_true",
                help="""Restores the original script file""")

    def restore_script(self, trial):
        if self.args_file:
            super(NowRestore, self).restore_script(trial)
        self.trial = trial

    def execute(self, func, line, cell, magic_cls):
        argv, args = self.arguments(func, line)
        self.args_file, self.trial = args.file, None
        super(NowRestore, self).do_restore(args)
        return persistence.get(self.trial.code_hash)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time

from copy import deepcopy
from collections import namedtuple, OrderedDict, defaultdict

from .model import Model
from .trial import Trial, TreeElement, Single, Call
from .trial import Mixed, Group
from .trial_activation_visitors import TrialGraphVisitor
from .trial_activation_visitors import TrialGraphCombineVisitor
from ..utils import OrderedCounter, concat_iter


class hashabledict(dict):
    def el(self, e):
        if isinstance(e, dict):
            return hashabledict(e)
        else:
            return e

    def __key(self):
        return tuple((k,self.el(self[k])) for k in sorted(self))
    def __hash__(self):
        return hash(self.__key())
    def __eq__(self, other):
        return self.__key() == other.__key()


class activationdict(dict):
    def __key(self):
        return tuple((k,self[k]) for k in sorted(self))
    def __hash__(self):
        return hash(self.__key())
    def __eq__(self, other):
        return self['name'] == other['name']


class fadict(dict):
    def __key(self):
        return (self['name'],
                self['content_hash_before'],
                self['content_hash_after'])
    def __hash__(self):
        return hash(self.__key())
    def __eq__(self, other):
        return ((self['content_hash_before'] == other['content_hash_before'])
            and (self['content_hash_after'] == other['content_hash_after']))


class Diff(Model):
    """ This model represents a diff between two trials
    Initialize it by passing both trials ids:
        diff = Diff(2)

    There are two visualization modes for the graph:
        exact match: calls are only combined when all the sub-call match
            diff.graph_type = 0
        combined: calls are combined without considering the sub-calls
            diff.graph_type = 1

    There are also three visualization modes for the diff:
        combine graphs: combines both trial graphs
            diff.display_mode = 0
        side by side: displays both graphs side by side
            diff.display_mode = 1
        combined and side by side: combine graphs and displays both separated graphs
            diff.display_mode = 2


    You can change the graph width and height by the variables:
        diff.graph_width = 600
        diff.graph_height = 400
    """

    DEFAULT = {
        'graph_width': 500,
        'graph_height': 500,
        'graph_type': 0,
        'display_mode': 0,
    }

    def __init__(self, trial_id1, trial_id2, exit=False, **kwargs):
        super(Diff, self).__init__(trial_id1, trial_id2, exit=exit, **kwargs)
        self.initialize_default(kwargs)

        self.trial1 = Trial(trial_id1, exit=exit)
        self.trial2 = Trial(trial_id2, exit=exit)
        self._graph_types = {
            0: self.independent_naive_activation_graph,
            1: self.combined_naive_activation_graph
        }
        self._display_modes = {
            0: "combined",
            1: "side by side",
            2: "both",
        }
        self._independent_cache = None
        self._combined_cache = None


    def trial(self):
        """ Returns a tuple with the information of both trials """
        return diff_dict(self.trial1.info(), self.trial2.info())

    def modules(self):
        """ Diffs modules from trials """
        fn = lambda x: hashabledict(x)
        return diff_set(
            set(self.trial1.modules(fn)[1]),
            set(self.trial2.modules(fn)[1]))

    def environment(self):
        """ Diffs environment variables """
        return diff_set(
            dict_to_set(self.trial1.environment()),
            dict_to_set(self.trial2.environment()))

    def file_accesses(self):
        """ Diffs file accesses """
        return diff_set(
            set(fadict(fa) for fa in self.trial1.file_accesses()),
            set(fadict(fa) for fa in self.trial2.file_accesses()))

    def independent_naive_activation_graph(self):
        """ Generates an activation graph for both trials and transforms it into an
            exact match graph supported by d3 """
        if not self._independent_cache:
            g1 = self.trial1.independent_activation_graph()
            g2 = self.trial2.independent_activation_graph()
            self._independent_cache = NaiveGraphDiff(g1, g2).to_dict(), g1, g2
        return self._independent_cache

    def combined_naive_activation_graph(self):
        """ Generates an activation graph for both trials and transforms it into an
            combined graph supported by d3 """
        if not self._combined_cache:
            g1 = self.trial1.combined_activation_graph()
            g2 = self.trial2.combined_activation_graph()
            self._combined_cache = NaiveGraphDiff(g1, g2).to_dict(), g1, g2
        return self._combined_cache

    def _repr_html_(self):
        """ Displays d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-diff" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id1="{id1}" data-id2="{id2}" data-mode="{mode}">
                {data}
            </div>
        """.format(
            uid=uid, id1=self.trial1.id, id2=self.trial2.id,
            mode=self.display_mode,
            data=self.escape_json(self._graph_types[self.graph_type]()),
            width=self.graph_width, height=self.graph_height)
        return result


class NaiveGraphDiff(object):

    def __init__(self, g1, g2):
        self.id = 0
        self.nodes = []
        self.edges = []
        self.context_edges = {}
        self.old_to_new = {}
        self.max_duration = dict(concat_iter(
            g1['max_duration'].items(), g2['max_duration'].items()))

        self.min_duration = dict(concat_iter(
            g1['min_duration'].items(), g2['min_duration'].items()))

        self.merge(g1, g2)

    def fix_caller_id(self, graph):
        called = {}
        seq = defaultdict(list)
        nodes = graph['nodes']
        edges = [hashabledict(x) for x in graph['edges']]
        for edge in edges:
            if edge['type'] == 'call':
                called[edge['target']] = edge['source']
            if edge['type'] == 'sequence':
                seq[edge['source']].append(edge)

        visited = set()
        while called:
            t = {}
            for nid, parent in called.items():
                nodes[nid]['caller_id'] = parent
                visited.add(nid)
                for e in seq[nid]:
                    if not e['target'] in visited:
                        t[e['target']] = parent
            called = t

    def merge(self, g1, g2):
        self.fix_caller_id(g1)
        self.fix_caller_id(g2)
        nodes1 = [hashabledict(x) for x in g1['nodes']]
        nodes2 = [hashabledict(x) for x in g2['nodes']]
        def cmp_node(x, y):
            if x['name'] != y['name']:
                return False
            if x['caller_id'] is None and y['caller_id'] is not None:
                return False
            if x['caller_id'] is not None and y['caller_id'] is None:
                return False
            if x['caller_id'] is None and y['caller_id'] is None:
                return True
            caller1, caller2 = nodes1[x['caller_id']], nodes2[y['caller_id']]
            return caller1['name'] == caller2['name']

        res, _ = lcs(nodes1, nodes2, cmp_node)

        for a, b in res.items():
            n = deepcopy(a)
            del n['node']
            n['node1'] = a['node']
            n['node2'] = b['node']
            n['node1']['original'] = a['index']
            n['node2']['original'] = b['index']
            n['index'] = self.id
            a['node']['diff'] = self.id
            b['node']['diff'] = self.id
            self.old_to_new[(1, a['index'])] = self.id
            self.old_to_new[(2, b['index'])] = self.id
            self.id += 1
            self.nodes.append(n)

        self.add_nodes(nodes1, 1)
        self.add_nodes(nodes2, 2)
        self.add_edges(g1['edges'], 1)
        self.add_edges(g2['edges'], 2)

    def add_nodes(self, nodes, ng):
        for node in nodes:
            if not (ng, node['index']) in self.old_to_new:
                nid = self.add_node(node)
                self.old_to_new[(ng, node['index'])] = nid
                node['node']['diff'] = nid

    def add_edges(self, edges, ng):
        for edge in edges:
            if (ng, edge['source']) in self.old_to_new:
                self.add_edge(self.old_to_new[(ng, edge['source'])],
                              self.old_to_new[(ng, edge['target'])],
                              edge)

    def add_node(self, node):
        n = deepcopy(node)
        node_id = n['index'] = self.id
        n['node']['original'] = node['index']
        self.nodes.append(n)
        self.id += 1
        return node_id

    def add_edge(self, source, target, edge):
        edge_key = "{} {} {}".format(source, target, edge['type'])

        if not edge_key in self.context_edges:
            e = deepcopy(edge)
            e['source'] = source
            e['target'] = target
            self.edges.append(e)
            self.context_edges[edge_key] = e
        else:
            e = self.context_edges[edge_key]
            e['count'] = (e['count'], edge['count'])

    def to_dict(self):
        return {
            'max_duration': self.max_duration,
            'min_duration': self.min_duration,
            'nodes': self.nodes,
            'edges': self.edges,
        }


def dict_to_set(d):
    result = set()
    for key, value in d.items():
        result.add(activationdict({'name': key, 'value': value}))
    return result

def diff_dict(before, after):
    result = {}
    for key in before.keys():
        if key != 'id' and before[key] != after[key]:
            result[key] = [before[key], after[key]]
    return result

def diff_set(before, after):
    removed = before - after
    added = after - before
    replaced = set()

    removed_by_name = {}
    for element_removed in removed:
        removed_by_name[element_removed['name']] = element_removed
    for element_added in added:
        element_removed = removed_by_name.get(element_added['name'])
        if element_removed:
            replaced.add((element_removed, element_added))
    for (element_removed, element_added) in replaced:
        removed.discard(element_removed)
        added.discard(element_added)

    return (added, removed, replaced)

def lcs(a, b, eq=lambda x, y: x == y):
    lengths = [[0 for j in range(len(b)+1)] for i in range(len(a)+1)]
    # row 0 and column 0 are initialized to 0 already
    for i, x in enumerate(a):
        for j, y in enumerate(b):
            if eq(x, y):
                lengths[i+1][j+1] = lengths[i][j] + 1
            else:
                lengths[i+1][j+1] = \
                    max(lengths[i+1][j], lengths[i][j+1])
    # read the substring out from the matrix
    result_a, result_b = OrderedCounter(), OrderedCounter()
    x, y = len(a), len(b)
    while x != 0 and y != 0:
        if lengths[x][y] == lengths[x-1][y]:
            x -= 1
        elif lengths[x][y] == lengths[x][y-1]:
            y -= 1
        else:
            result_a[a[x-1]] = b[y-1]
            result_b[b[y-1]] = a[x-1]
            x -= 1
            y -= 1
    return result_a, result_b



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

# Do not add from __future__ imports here
"""Provide support for both Python 2 and Python 3"""

import sys


try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO


def cross_compile(*args, **kwargs):
    """Compile the source string into a code object

    __future__ imports change the behavior of default compile function
    This function just provides the 'compile' free of __future__ imports
    """
    return compile(*args, **kwargs)


def bytes_string(text, encode='utf-8'):
    """Return a bytes object on Python 3 and a str object on Python 2"""
    if sys.version_info < (3, 0):
        if isinstance(text, unicode):
            result = text.encode(encode)
        else:
            result = text
    else:
        if isinstance(text, bytes):
            result = text
        else:
            result = bytes(text, encode)
    return result



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


def load_ipython_extension(ipython):
    from .now.ipython import init
    init(ipython=ipython)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from collections import defaultdict

from IPython.core.magic import Magics, magics_class

from .now_run import NowRun
from .now_ipython import NowIpython
from .now_set_default import NowSetDefault
from .now_sql import NowSQL
from .now_prolog import NowProlog
from .now_restore import NowRestore
from .now_ls_magic import NowLsMagic


MAGICS = [
	('now_run', 'line_cell', NowRun),
	('now_ip', 'line', NowIpython),
	('now_set_default', 'line', NowSetDefault),
    ('now_sql', 'cell', NowSQL),
    ('now_prolog', 'cell', NowProlog),
    ('now_restore', 'line', NowRestore),
	('now_ls_magic', 'line', NowLsMagic),
]


@magics_class
class NoworkflowMagics(Magics):

    def __init__(self, shell):
        super(NoworkflowMagics, self).__init__(shell=shell)
        self.commands = [
        	cls(command, cls.__doc__, magic_type=magic_type)
        	for command, magic_type, cls in MAGICS
        ]
        self.now_magics = defaultdict(dict)
        self._generate_magics()

    def _generate_magics(self):
        for command in self.commands:
            command.add_arguments()
            def func(line, cell=None, c=command):
                return c.execute(c.func, line, cell, self)

            command.func = command.create_magic(func)
            for typ in command.magic_type.split('_'):
                self.magics[typ][command.magic] = command.func
                self.now_magics[typ][command.magic] = command.func


def register_magics(ipython):
    if not ipython:
        ipython = get_ipython()
    magics = NoworkflowMagics(ipython)
    ipython.register_magics(magics)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
from .profiler import Profiler, InspectProfiler
from .slicing import Tracer

provider = None

def provenance_provider(execution_provenance):
    glob = globals()
    if execution_provenance in glob:
        return glob[execution_provenance]
    return Profiler

def enable(args, metascript):
    global provider
    provider = provenance_provider(args.execution_provenance)(
        metascript, args.depth_context, args.depth,
    )
    provider.tearup()


def disable():
    global provider
    provider.teardown()


def store():
    global provider
    provider.store()
# TODO: Processor load. Should be collected from time to time (there are static and dynamic metadata)
# print os.getloadavg()


# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import ast
import sys
import bisect

from collections import defaultdict

from ..utils import print_msg
from .function_visitor import FunctionVisitor
from .context import NamedContext
from .utils import (FunctionCall, ClassDef, Decorator, Generator, Assert,
                    index, safeget)


class AssignLeftVisitor(ast.NodeVisitor):

    def __init__(self):
        self.names = []
        self.enable = True
        self.last = ""

    def visit_Attribute(self, node):
        self.generic_visit(node)
        if self.enable:
            self.last += '.' + node.attr
            self.names.append((self.last, node.ctx, node.lineno))

    def visit_Subscript(self, node):
        self.visit(node.value)
        self.enable = False
        self.visit(node.slice)

    def visit_Name(self, node):
        if self.enable:
            self.last = node.id
            self.names.append((self.last, node.ctx, node.lineno))
        self.generic_visit(node)


class AssignRightVisitor(ast.NodeVisitor):

    def __init__(self):
        self.names = []
        self.special = NamedContext()
        self.line = -1

    def add(self, name, ctx, lineno):
        if not self.special.use:
            self.names.append((name, ctx, lineno))
        else:
            self.special.add(name)

    def in_special(self, node):
        return node.id in self.special.flat()

    def max_line(self, node):
        try:
            self.line = max(node.lineno, self.line)
        except:
            pass

    def visit_Name(self, node):
        self.max_line(node)
        if not self.in_special(node):
            self.add(node.id, node.ctx, node.lineno)
        self.generic_visit(node)

    def visit_Lambda(self, node):
        self.max_line(node)
        self.special.enable()
        self.visit(node.args)
        self.special.disable()
        self.visit(node.body)
        self.special.pop()

    def visit_ListComp(self, node):
        self.max_line(node)
        self.special.enable()
        self.special.disable()
        for gen in node.generators:
            self.visit(gen)
        self.visit(node.elt)
        self.special.pop()

    def visit_SetComp(self, node):
        self.max_line(node)
        self.visit_ListComp(node)

    def visit_GeneratorExp(self, node):
        self.max_line(node)
        self.visit_ListComp(node)

    def visit_DictComp(self, node):
        self.max_line(node)
        self.special.enable()
        self.special.disable()
        for gen in node.generators:
            self.visit(gen)
        self.visit(node.key)
        self.visit(node.value)
        self.special.pop()

    def visit_comprehension(self, node):
        self.max_line(node)
        self.special.use = True
        self.visit(node.target)
        self.special.disable()
        self.visit(node.iter)
        for _if in node.ifs:
            self.visit(_if)

    def visit_Call(self, node):
        self.max_line(node)
        self.add(node.uid, 'fn', node.lineno)


def tuple_or_list(node):
    return isinstance(node, ast.Tuple) or isinstance(node, ast.List)


def add_all_names(dest, origin):
    for name in origin:
        dest.append(name)


def assign_dependencies(target, value, dependencies, conditions, loop,
                        aug=False, testlist_star_expr=True):
    left, right = AssignLeftVisitor(), AssignRightVisitor()

    if testlist_star_expr and tuple_or_list(target) and tuple_or_list(value):
        for i, targ in enumerate(target.elts):
            assign_dependencies(targ, value.elts[i], dependencies, conditions,
                                loop, testlist_star_expr=True)
        return

    left.visit(target)
    if value:
        right.visit(value)

    for name, ctx, lineno in left.names:
        lineno = right.line if right.line != -1 else lineno
        self_reference = False
        dependencies[lineno][name]
        for value, ctx2, lineno2 in right.names:
            dependencies[lineno][name].append(value)
            if name == value:
                self_reference = True

        if aug:
            dependencies[lineno][name].append(name)
            self_reference = True

        if self_reference:
            add_all_names(dependencies[lineno][name], loop)

        add_all_names(dependencies[lineno][name], conditions)


class SlicingVisitor(FunctionVisitor):

    def __init__(self, *args):
        super(SlicingVisitor, self).__init__(*args)
        self.path = self.metascript['path']
        self.name_refs = defaultdict(lambda: {
            'Load': [], 'Store': [], 'Del': [],
            'AugLoad': [], 'AugStore': [], 'Param': [],
        })
        self.dependencies = defaultdict(lambda: defaultdict(list))
        self.function_calls = defaultdict(dict)
        self.function_calls_by_lasti = defaultdict(dict)
        self.function_calls_list = []
        self.imports = set()
        self.condition = NamedContext()
        self.loop = NamedContext()

    def add_call_function(self, node, cls, *args):
        function_call = cls(AssignRightVisitor, *args)
        function_call.visit(node)
        function_call.line, function_call.col = node.uid
        self.function_calls_list.append(function_call)
        return function_call

    def add_decorator(self, node):
        self.add_call_function(node, Decorator)

    def add_generator(self, typ, node):
        self.add_call_function(node, Generator, typ)

    def visit_stmts(self, stmts):
        for stmt in stmts:
            self.visit(stmt)

    def visit_AugAssign(self, node):
        assign_dependencies(node.target, node.value,
                            self.dependencies,
                            self.condition.flat(),
                            self.loop.flat(), aug=True,
                            testlist_star_expr=False)
        self.generic_visit(node)

    def visit_Assign(self, node):
        for target in node.targets:
            assign_dependencies(target, node.value,
                                self.dependencies,
                                self.condition.flat(),
                                self.loop.flat())

        self.generic_visit(node)

    def visit_For(self, node):
        assign_dependencies(node.target, node.iter,
                            self.dependencies,
                            self.condition.flat(),
                            self.loop.flat(),
                            testlist_star_expr=False)
        self.loop.enable()
        self.visit(node.target)
        self.loop.disable()
        self.visit(node.iter)
        self.visit_stmts(node.body)
        self.visit_stmts(node.orelse)
        self.loop.pop()

    def visit_While(self, node):
        self.condition.enable()
        self.visit(node.test)
        self.condition.disable()
        self.visit_stmts(node.body)
        self.visit_stmts(node.orelse)
        self.condition.pop()

    def visit_If(self, node):
        self.visit_While(node)

    def visit_Name(self, node):
        self.condition.add(node.id)
        self.loop.add(node.id)
        self.name_refs[node.lineno][type(node.ctx).__name__]\
            .append(node.id)
        self.generic_visit(node)

    def visit_Call(self, node):
        self.call(node)
        self.generic_visit(node)
        fn = self.add_call_function(node, FunctionCall)
        self.function_calls[fn.line][fn.col] = fn

    def visit_Return(self, node):
        assign_dependencies(ast.Name('return', ast.Store(),
                                     lineno=node.lineno),
                            node.value,
                            self.dependencies,
                            self.condition.flat(),
                            self.loop.flat(),
                            testlist_star_expr=False)
        if node.value:
            self.visit(node.value)

    def visit_Yield(self, node):
        self.visit_Return(node)


    def visit_ClassDef(self, node):
        self.generic_visit(node)
        self.add_call_function(node, ClassDef)

        for dec_node in node.decorator_list:
            self.add_decorator(dec_node)

    def visit_ListComp(self, node):
        self.generic_visit(node)
        if sys.version_info >= (3, 0):
            for gen_node in node.generators:
                self.add_generator('List', gen_node)

    def visit_SetComp(self, node):
        self.generic_visit(node)
        for gen_node in node.generators:
            self.add_generator('Set', gen_node)

    def visit_DictComp(self, node):
        self.generic_visit(node)
        for gen_node in node.generators:
            self.add_generator('Dict', gen_node)

    def visit_GeneratorExp(self, node):
        self.generic_visit(node)
        for gen_node in node.generators:
            self.add_generator('Generator', gen_node)

    def visit_FunctionDef(self, node):
        self.generic_visit(node)

        for dec_node in node.decorator_list:
            self.add_decorator(dec_node)

    def visit_Assert(self, node):
        self.generic_visit(node)
        if node.msg:
            self.add_call_function(node, Assert, node.msg)

        # ToDo: with msg self.function_calls_list.append(cls)

    def teardown(self):
        """Matches AST call order to call order in disassembly
        Possible issues:
        1- The disassembly may be specific to cpython. It may not work on other
        implementations
        2- If the order in AST is not correct, the matching will fail
        3- If there are other CALL_FUNCTION that are not considered in the AST
        the matching will fail
            both visit_ClassDef and visit_Call generates CALL_FUNCTION
        """
        output = self.disasm
        self.disasm = []

        line = -1
        col = 0
        for disasm in output:
            num = disasm[:8].strip()
            if num:
                line = int(num)
            splitted = disasm.split()
            i = index(splitted, ('CALL_FUNCTION', 'CALL_FUNCTION_VAR',
                                 'CALL_FUNCTION_KW', 'CALL_FUNCTION_VAR_KW'))

            if not i is None:
                f_lasti = int(splitted[i-1])
                call = safeget(self.function_calls_list, col)
                call.lasti = f_lasti
                self.function_calls_by_lasti[line][f_lasti] = call

                col += 1
                self.disasm.append(
                    "{} | {}".format(disasm, call))
            else:
                self.disasm.append(disasm)

            if not index(splitted, ('IMPORT_NAME', 'IMPORT_FROM')) is None:
                self.imports.add(line)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from datetime import datetime


FORMAT = '%Y-%m-%d %H:%M:%S.%f'


def calculate_duration(obj):
    return int((
        datetime.strptime(obj['finish'], FORMAT) -
        datetime.strptime(obj['start'], FORMAT)
    ).total_seconds() * 1000000)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


class Definition(object):

	def __init__(self, metascript):
		self.metascript = metascript

		self.paths = []
		# Map of dependencies by line
		self.line_dependencies = {}
		# Map of name_refs by line
		self.line_usages = {}
		# Map of calls by line and col
		self.call_by_col = {}
		# Map of calls by offset line and lasti
		self.call_by_lasti = {}
		# Set of imports
		self.imports = {}

	def add_visitor(self, visitor):
		self.paths.append(visitor.path)
		self.line_dependencies[visitor.path] = visitor.dependencies
		self.line_usages[visitor.path] = visitor.name_refs
		self.call_by_col[visitor.path] = visitor.function_calls
		self.call_by_lasti[visitor.path] = visitor.function_calls_by_lasti
		self.imports[visitor.path] = visitor.imports



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse

from .command import Command
from .cmd_run import Run, LAST_TRIAL
from .cmd_list import List
from .cmd_show import Show
from .cmd_diff import Diff
from .cmd_export import Export
from .cmd_restore import Restore
from .cmd_vis import Vis

def main():

    parser = argparse.ArgumentParser(description = __doc__)
    subparsers = parser.add_subparsers()
    commands = [
        Run('runs a script collecting its provenance'),
        List('lists all trials registered in the current directory'),
        Show('shows the collected provenance of a trial'),
        Diff('compares the collected provenance of two trials'),
        Export('exports the collected provenance of a trial to Prolog'),
        Restore('restore the files of a trial'),
        Vis('visualization tool'),
    ]
    for cmd in commands:
        cmd.create_parser(subparsers)

    args, _ = parser.parse_known_args()
    args.func(args)

__all__ = [
    b'LAST_TRIAL',
    b'Command',
    b'Run',
    b'List',
    b'Show',
    b'Diff',
    b'Export',
    b'Restore',
    b'Vis',
    b'main',
]


# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import sys
import os
import itertools
import inspect
from datetime import datetime
from .base import StoreOpenMixin
from .activation import Activation

from ..persistence import persistence

class Profiler(StoreOpenMixin):

    def __init__(self, *args):
        super(Profiler, self).__init__(*args)
        # the number of user functions activated
        #   (starts with -1 to compensate the first call to the script itself)
        self.depth_user = -1
        # the number of non-user functions activated
        self.depth_non_user = 0
        self.activation_stack = []
        self.function_activation = None

        self.event_map['c_call'] = self.trace_c_call
        self.event_map['call'] = self.trace_call
        self.event_map['c_return'] = self.trace_c_return
        self.event_map['c_exception'] = self.trace_c_exception
        self.event_map['return'] = self.trace_return

    def add_file_access(self, file_access):
        self.activation_stack[-1].file_accesses.append(file_access)

    def valid_depth(self):
        depth = self.depth_user + self.depth_non_user
        valid_all_threshold = depth <= self.depth_threshold
        valid_non_user_threshold = self.depth_non_user <= self.depth_threshold
        return ((self.depth_context == 'all' and valid_threshold) or
               (self.depth_context == 'non-user' and valid_non_user_threshold))

    def add_activation(self, activation):
        if activation:
            activation.start = datetime.now()
            self.activation_stack.append(activation)

    def close_activation(self, event, arg):
        activation = self.activation_stack.pop()
        activation.finish = datetime.now()
        try:
            activation.return_value = repr(arg) if event == 'return' else None
        except:  # ignoring any exception during capture
            activation.return_value = None
        # Update content of accessed files
        for file_access in activation.file_accesses:
            # Checks if file still exists
            if os.path.exists(file_access['name']):
                with persistence.std_open(file_access['name'], 'rb') as f:
                    file_access['content_hash_after'] = persistence.put(
                        f.read())

        if self.activation_stack:
            # Store the current activation in the previous activation
            self.activation_stack[-1].function_activations.append(activation)
        else:
            # Store the current activation as the first activation
            self.function_activation = activation
            self.enabled = False

    def trace_c_call(self, frame, event, arg):
        self.depth_non_user += 1
        if self.valid_depth():
            self.add_activation(Activation(
                arg.__name__ if arg.__self__ == None else '.'.join(
                    [type(arg.__self__).__name__, arg.__name__]),
                frame.f_lineno, frame.f_lasti
            ))

    def capture_python_params(self, frame, activation):
        values = frame.f_locals
        co = frame.f_code
        names = co.co_varnames
        nargs = co.co_argcount
        # Capture args
        for var in itertools.islice(names, 0, nargs):
            try:
                activation.arguments[var] = repr(values[var])
                activation.args.append(var)
            except:
                # ignoring any exception during capture
                pass
        # Capture *args
        if co.co_flags & inspect.CO_VARARGS:
            varargs = names[nargs]
            activation.arguments[varargs] = repr(values[varargs])
            activation.starargs.append(varargs)
            nargs += 1
        # Capture **kwargs
        if co.co_flags & inspect.CO_VARKEYWORDS:
            kwargs = values[names[nargs]]
            for key in kwargs:
                activation.arguments[key] = repr(kwargs[key])
            activation.kwargs.append(names[nargs])

    def trace_call(self, frame, event, arg):
        #print("now", frame.f_back.f_lineno, frame.f_code.co_name)
        co_name = frame.f_code.co_name
        co_filename = frame.f_code.co_filename
        if self.script == co_filename:
            self.depth_user += 1
        else:
            self.depth_non_user += 1

        if self.valid_depth():
            activation =  Activation(
                co_name if co_name != '<module>' else co_filename,
                frame.f_back.f_lineno, frame.f_back.f_lasti
            )
            # Capturing arguments
            self.capture_python_params(frame, activation)

            # Capturing globals
            function_def = persistence.load(
                'function_def',
                name = repr(activation.name),
                trial_id = self.trial_id
            ).fetchone()

            if function_def:
                global_vars = persistence.load(
                    'object',
                    type='"GLOBAL"',
                    function_def_id = function_def[str('id')]
                )
                aglobals = activation.globals
                fglobals = frame.f_globals
                for global_var in global_vars:
                    aglobals[global_var['name']] = fglobals[global_var['name']]

            self.add_activation(activation)

    def trace_c_return(self, frame, event, arg):
        if self.valid_depth():
            self.close_activation(event, arg)
        self.depth_non_user -= 1

    def trace_c_exception(self, frame, event, arg):
        if self.valid_depth():
            self.close_activation(event, arg)
        self.depth_non_user -= 1

    def trace_return(self, frame, event, arg):
        if self.valid_depth():
            self.close_activation(event, arg)

        if self.script == frame.f_code.co_filename:
            self.depth_user -= 1
        else:
            self.depth_non_user -= 1

    def tracer(self, frame, event, arg):
        # Only enable activations gathering after the first call to the script
        co_filename = frame.f_code.co_filename
        if not self.enabled and event == 'call' and self.script == co_filename:
            self.enabled = True

        if self.enabled:
            super(Profiler, self).tracer(frame, event, arg)
        return self.tracer

    def store(self):
        now = datetime.now()
        persistence.update_trial(self.trial_id, now, self.function_activation)

    def tearup(self):
        sys.setprofile(self.tracer)


class InspectProfiler(Profiler):
    """ This Profiler uses the inspect.getargvalues that is slower because
    it considers the existence of anonymous tuple """

    def capture_python_params(self, frame, activation):
        (args, varargs, keywords, values) = inspect.getargvalues(frame)
        for arg in args:
            try:
                activation.arguments[arg] = repr(values[arg])
                activation.args.append(arg)
            except:  # ignoring any exception during capture
                pass
        if varargs:
            activation.arguments[varargs] = repr(values[varargs])
            activation.starargs.append(varargs)
        if keywords:
            for key, value in values[keywords].iteritems():
                activation.arguments[key] = repr(value)
                activation.kwargs.append(key)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import re
import argparse

from ...formatter import PrettyLines
from ..models import set_default
from .command import IpythonCommandMagic


class NowLsMagic(IpythonCommandMagic):
    """Returns the list of noWorkflow magics"""

    def execute(self, func, line, cell, magic_cls):
        cline, ccell = '%', '%%'
        magics = magic_cls.now_magics
        out = ['Line magics:',
               cline + ('  ' + cline).join(sorted(magics['line'])),
               '',
               'Cell magics:',
               ccell + ('  ' + ccell).join(sorted(magics['cell']))]
        return PrettyLines(out)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define utility functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from textwrap import dedent
from collections import OrderedDict, Counter
from pkg_resources import resource_string


LABEL = '[now] '
verbose = False


def wrap(string, initial="  ", other="\n  "):
    """Re-indent indented text"""
    return initial + other.join(dedent(string).split('\n'))


def print_msg(message, force=False):
    """Print message with [now] prefix when in verbose mode"""
    if verbose or force:
        print('{}{}'.format(LABEL, message))


def print_fn_msg(message, force=False):
    """Print lazy message with [now] prefix"""
    if verbose or force:
        print('{}{}'.format(LABEL, message()))


def print_map(title, a_map):
    """Print map"""
    print_msg(title, True)
    output = []
    for key in a_map:
        output.append('  {}: {}'.format(key, a_map[key]))
    print('\n'.join(sorted(output)))


def print_modules(modules):
    """Print modules"""
    output = []
    for module in modules:
        output.append(wrap('''\
            Name: {name}
            Version: {version}
            Path: {path}
            Code hash: {code_hash}\
            '''.format(**module), other="\n    "))
    print('\n\n'.join(output))


def print_environment_attrs(environment_attrs):
    """Print environment variables"""
    output = []
    for environment_attr in environment_attrs:
        output.append('  {name}: {value}'.format(**environment_attr))
    print('\n'.join(output))


def resource(filename, encoding=None):
    """Access resource content via setuptools"""
    content = resource_string(__name__, filename)
    if encoding:
        return content.decode(encoding=encoding)
    return content


class OrderedCounter(OrderedDict, Counter):
    """OrderedDict with default value 0"""

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__,
                           OrderedDict(self))
    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)


def concat_iter(*iters):
    """Concatenate iterators"""
    for iterator in iters:
        for value in iterator:
            yield value



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import time
from datetime import datetime
from collections import defaultdict, OrderedDict, Counter
from pyposast.cross_version import buffered_str


from .. import utils
from ..formatter import PrettyLines
from ..persistence import row_to_dict, persistence
from .model import Model
from .trial_activation_visitors import TrialGraphVisitor
from .trial_activation_visitors import TrialGraphCombineVisitor
from .trial_prolog import TrialProlog
from .utils import calculate_duration, FORMAT
from .activation import Activation


class Trial(Model):
    """ This model represents a trial
    Initialize it by passing the trial id:
        trial = Trial(2)

    There are two visualization modes for the graph:
        exact match: calls are only combined when all the sub-call match
            trial.graph_type = 0
        combined: calls are combined without considering the sub-calls
            trial.graph_type = 1

    You can change the graph width and height by the variables:
        trial.graph_width = 600
        trial.graph_height = 400
    """

    DEFAULT = {
        'graph_width': 500,
        'graph_height': 500,
        'graph_type': 0,
    }

    def __init__(self, trial_id, exit=False, script=None, **kwargs):
        super(Trial, self).__init__(trial_id, exit=exit, script=script,
                                    **kwargs)
        self.initialize_default(kwargs)

        if exit:
            last_trial_id = persistence.last_trial_id(script=script)
            trial_id = trial_id or last_trial_id
            if not 1 <= trial_id <= last_trial_id:
                utils.print_msg('inexistent trial id', True)
                sys.exit(1)

        self.id = trial_id
        self._info = None
        self._graph_types = {
            0: self.independent_activation_graph,
            1: self.combined_activation_graph
        }
        self.trial_prolog = TrialProlog(self)

    def query(self, query):
        return self.trial_prolog.query(query)

    def prolog_rules(self):
        return self.trial_prolog.export_rules().split('\n')

    @property
    def trial_id(self):
        from warnings import warn
        warn('trial_id propery deprecated. Please use id')
        return self.id

    @property
    def script(self):
        """ Returns the "main" script of the trial """
        info = self.info()
        return info['script']

    @property
    def script_content(self):
        """ Returns the "main" script content of the trial """
        return PrettyLines(
            buffered_str(persistence.get(self.code_hash)).split('/n'))

    @property
    def code_hash(self):
        """ Returns the hash code of the main script """
        info = self.info()
        return info['code_hash']

    def _repr_html_(self):
        """ Displays d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-trial" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id="{id}">
                {data}
            </div>
        """.format(
            uid=uid, id=self.id,
            data=self.escape_json(self._graph_types[self.graph_type]()),
            width=self.graph_width, height=self.graph_height)
        return result

    def info(self):
        """ Returns dict with the trial information, considering the duration """
        if self._info is None:
            self._info = row_to_dict(
                persistence.load_trial(self.id).fetchone())
            if self._info['finish']:
                self._info['duration'] = calculate_duration(self._info)
        return self._info

    def function_defs(self):
        """ Returns a dict of function definitions """
        return {
            function['name']: row_to_dict(function)
            for function in persistence.load('function_def',
                                             trial_id=self.id)
        }

    def head_trial(self, remove=False):
        """ Returns the parent trial object """
        parent_id = persistence.load_parent_id(self.script, remove=remove)
        return Trial(parent_id)

    def modules(self, map_fn=row_to_dict):
        """ Returns the modules imported during the trial
            The first element is a list of local modules
            The second element is a list of external modules
        """
        dependencies = persistence.load_dependencies(self.id)
        result = list(map(map_fn, dependencies))
        local = [dep for dep in result
                 if dep['path'] and persistence.base_path in dep['path']]
        return local, result

    def environment(self):
        """ Returns a dict of environment variables """
        return {
            attr['name']: attr['value'] for attr in map(row_to_dict,
                persistence.load('environment_attr', trial_id=self.id))
        }

    def file_accesses(self):
        """ Returns a list of file accesses """
        file_accesses = persistence.load('file_access',
                                         trial_id=self.id)

        result = []
        for fa in map(row_to_dict, file_accesses):
            stack = []
            function_activation = next(iter(self.activations(
                id=fa['function_activation_id'])))
            while function_activation:
                function_name = function_activation['name']
                try:
                    function_activation = next(iter(self.activations(
                        id=function_activation['caller_id'])))
                    stack.insert(0, function_name)
                except StopIteration:
                    function_activation = None
            if not stack or stack[-1] != 'open':
                stack.append(' ... -> open')

            result.append({
                'id': fa['id'],
                'function_activation_id': fa['function_activation_id'],
                'name': fa['name'],
                'mode': fa['mode'],
                'buffering': fa['buffering'],
                'content_hash_before': fa['content_hash_before'],
                'content_hash_after': fa['content_hash_after'],
                'timestamp': fa['timestamp'],
                'stack': ' -> '.join(stack),
            })
        return result

    def activations(self, **conditions):
        """ Returns a list of activations """
        return map(Activation, persistence.load('function_activation',
                                                trial_id=self.id,
                                                order='start',
                                                **conditions))

    def slicing_variables(self):
        """ Returns a list of slicing variables """
        return persistence.load('slicing_variable',
                                trial_id=self.id,
                                order='vid ASC')

    def slicing_usages(self):
        """ Returns a list of slicing usages """
        return persistence.load('slicing_usage',
                                trial_id=self.id)

    def slicing_dependencies(self):
        """ Returns a list of slicing dependencies """
        return persistence.load('slicing_dependency',
                                trial_id=self.id)

    def activation_graph(self):
        """ Generates an activation graph """
        result_stack = []
        stack = [Single(act) for act in self.activations()]

        if not stack:
            return TreeElement()

        result_stack.append(stack.pop())
        while stack:
            next = result_stack.pop()
            previous = stack.pop()
            add_flow(stack, result_stack, previous, next)

        return result_stack.pop()

    def independent_activation_graph(self):
        """ Generates an activation graph and transforms it into an
            exact match graph supported by d3 """
        graph = self.activation_graph()
        visitor = TrialGraphVisitor()
        graph.visit(visitor)
        return visitor.to_dict()

    def combined_activation_graph(self):
        """ Generates an activation graph and transforms it into an
            combined graph supported by d3 """
        graph = self.activation_graph()
        visitor = TrialGraphCombineVisitor()
        graph.visit(visitor)
        return visitor.to_dict()


class TreeElement(object):

    def __init__(self):
        self.duration = 0
        self.count = 1
        self.repr = ""

    def mean(self):
        if isinstance(self.duration, tuple):
            return (self.a.duration / self.a.count,
                    self.b.duration / self.b.count)
        return self.duration / self.count

    def visit(self, visitor):
        return visitor.visit_default(self)

    def calculate_repr(self):
        pass

    def mix(self, other):
        pass

    def __hash__(self):
        #return id(self)
        return hash(self.__repr__())

    def __repr__(self):
        return self.repr


class Single(TreeElement):

    def __init__(self, activation):
        self.activation = activation
        self.activations = {activation}
        self.parent = activation['caller_id']
        self.id = activation['id']
        self.line = activation['line']
        self.name = activation['name']
        self.trial_id = activation['trial_id']
        self.repr = "S({0}-{1})".format(self.line, self.name)

    @property
    def count(self):
        return sum(1 for a in self.activations)

    @count.setter
    def count(self, value):
        pass

    @property
    def duration(self):
        return sum(calculate_duration(a) for a in self.activations
                   if a['finish'] and a['start'])

    @duration.setter
    def duration(self, value):
        pass

    def mix(self, other):
        self.count += other.count
        self.duration += other.duration
        self.activations = self.activations.union(other.activations)

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if self.line != other.line:
            return False
        if self.name != other.name:
            return False
        return True

    def name_id(self):
        return "{0} {1}".format(self.line, self.name)

    def visit(self, visitor):
        return visitor.visit_single(self)

    def to_dict(self, nid):
        return {
            'index': nid,
            'caller_id': self.parent,
            'name': self.name,
            'node': {
                'trial_id': self.trial_id,
                'line': self.line,
                'count': self.count,
                'duration': self.duration,
                'info': Info(self)
            }
        }

    def __hash__(self):
        return super(Single, self).__hash__()


class Mixed(TreeElement):

    def __init__(self, activation):
        self.duration = activation.duration
        self.elements = [activation]
        self.parent = activation.parent
        self.id = activation.id
        self.repr = activation.repr

    @property
    def count(self):
        return sum(e.count for e in self.elements)

    @count.setter
    def count(self, value):
        pass

    @property
    def duration(self):
        return sum(e.duration for e in self.elements)

    @property
    def first(self):
        return next(iter(self.elements))

    @duration.setter
    def duration(self, value):
        pass

    def add_element(self, element):
        self.elements.append(element)

    def visit(self, visitor):
        return visitor.visit_mixed(self)

    def mix(self, other):
        self.elements += other.elements
        self.mix_results()

    def mix_results(self):
        it = iter(self.elements)
        initial = next(it)
        for element in it:
            initial.mix(element)

    def __hash__(self):
        return super(Mixed, self).__hash__()


class Group(TreeElement):

    def __init__(self):
        self.nodes = OrderedDict()
        self.edges = OrderedDict()
        self.duration = 0
        self.parent = None
        self.count = 1
        self.repr = ""

    def initialize(self, previous, next):
        self.nodes[next] = Mixed(next)
        self.duration = next.duration
        self.next = next
        self.last = next
        self.add_subelement(previous)
        self.parent = next.parent
        return self

    def add_subelement(self, previous):
        next, self.next = self.next, previous
        if not previous in self.edges:
            self.edges[previous] = utils.OrderedCounter()
        if not previous in self.nodes:
            self.nodes[previous] = Mixed(previous)
        else:
            self.nodes[previous].add_element(previous)
        self.edges[previous][next] += 1

    def calculate_repr(self):
        result = [
            "[{0}-{1}->{2}]".format(previous, count, next)
            for previous, edges in self.edges.items()
            for next, count in edges.items()
        ]

        self.repr = "G({0})".format(', '.join(result))

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if not self.edges == other.edges:
            return False
        return True

    def visit(self, visitor):
        return visitor.visit_group(self)

    def mix(self, other):
        for node, value in self.nodes.items():
            value.mix(other.nodes[node])

    def __hash__(self):
        return super(Group, self).__hash__()


class Call(TreeElement):

    def __init__(self, caller, called):
        self.caller = caller
        self.called = called
        self.called.calculate_repr()
        self.parent = caller.parent
        self.count = 1
        self.id = self.caller.id
        self.duration = self.caller.duration
        self.repr = 'C({0}, {1})'.format(self.caller, self.called)

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if not self.caller == other.caller:
            return False
        if not self.called == other.called:
            return False
        return True

    def visit(self, visitor):
        return visitor.visit_call(self)

    def mix(self, other):
        self.caller.mix(other.caller)
        self.called.mix(other.called)

    def __hash__(self):
        return super(Call, self).__hash__()


class Info(object):

    def __init__(self, single):
        self.title = ("Trial {trial}<br>"
                      "Function <b>{name}</b> called at line {line}").format(
            trial=single.trial_id, name=single.name, line=single.line)
        self.activations = set()
        self.duration = ""
        self.mean = ""
        self.extract_activations(single)

    def update_by_node(self, node):
        self.duration = self.duration_text(node['duration'], node['count'])
        self.mean = self.mean_text(node['mean'])
        self.activation_list = sorted(self.activations, key=lambda a: a[0])

    def add_activation(self, activation):
        self.activations.add(
            (datetime.strptime(activation['start'], FORMAT), activation))

    def extract_activations(self, single):
        for activation in single.activations:
            self.add_activation(activation)

    def duration_text(self, duration, count):
        return "Total duration: {} microseconds for {} activations".format(
            duration, count)

    def mean_text(self, mean):
        return "Mean: {} microseconds per activation".format(mean)

    def activation_text(self, activation):
        values = map(row_to_dict, persistence.load('object_value',
            function_activation_id=activation['id'], order='id'))
        values = [value for value in values if value['type'] == 'ARGUMENT']
        result = [
            "",
            "Activation #{id} from {start} to {finish} ({dur} microseconds)"
                .format(dur=calculate_duration(activation), **activation),
        ]
        if values:
            result.append("Arguments: {}".format(
                ", ".join("{}={}".format(value["name"], value["value"])
                    for value in values)))
        return result + [
            "Returned {}".format(activation['return'])
        ]

    def __repr__(self):
        result = [self.title, self.duration, self.mean]
        for activation in self.activation_list:
            result += self.activation_text(activation[1])

        return '<br/>'.join(result)


def join(a, b):
    if a == b:
        return Dual(a, b)
    return Branch(a, b)


def sequence(previous, next):
    if isinstance(next, Group):
        next.add_subelement(previous)
        return next
    return Group().initialize(previous, next)


def add_flow(stack, result, previous, next):
    if previous.parent == next.parent:
        # Same function level
        result.append(sequence(previous, next))

    elif previous.id == next.parent:
        # Previously called next
        # if top of result is in the same level of call:
        #   create sequece or combine results
        # if top of result is in a higher level, put Call on top of pile
        if result:
            add_flow(stack, result, Call(previous, next), result.pop())
        else:
            result.append(Call(previous, next))
    else:
        # Next is in a higher level
        # Put previous on top of result
        result.append(next)
        result.append(previous)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .model import Model
from .history import History
from .trial import Trial
from .diff import Diff
from .trial_prolog import TrialProlog



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sqlite3
import sys

from os.path import join, isdir, exists
from os import makedirs
from pkg_resources import resource_string
from collections import OrderedDict

from ..utils import print_msg, resource


PROVENANCE_DIRNAME = '.noworkflow'
CONTENT_DIRNAME = 'content'
DB_FILENAME = 'db.sqlite'
DB_SCRIPT = '../resources/noworkflow.sql'
PARENT_TRIAL = '.parent_config.json'


def row_to_dict(row):
    return OrderedDict(zip(row.keys(), row))


class Provider(object):

    def __init__(self, path=None, connect=False):
        self.base_path = None # Exeution path
        self.provenance_path = None  # Base .noworkflow path
        self.content_path = None # Base path for storing content of files
        self.parent_config_path = None # Base path for restore references
        self.db_conn = None # Connection to the database

        self.std_open = open # Original Python open function.

        if path:
            self.path = path

        if connect:
            self.connect(path)

    @property
    def path(self):
        return self.base_path

    @path.setter
    def path(self, path):
        self.base_path = path
        self.provenance_path = join(path, PROVENANCE_DIRNAME)
        self.content_path = join(self.provenance_path, CONTENT_DIRNAME)
        self.parent_config_path = join(self.provenance_path, PARENT_TRIAL)

    def connect(self, path=None):
        if path:
            self.path = path

        db_path = join(self.provenance_path, DB_FILENAME)

        if not isdir(self.content_path):
            makedirs(self.content_path)

        new_db = not exists(db_path)
        self.db_conn = sqlite3.connect(db_path)
        self.db_conn.row_factory = sqlite3.Row

        if new_db:
            print_msg('creating provenance database')
            # Accessing the content of a file via setuptools
            with self.db_conn as db:
                db.executescript(resource(DB_SCRIPT, 'utf-8'))

    def has_provenance(self, path=None):
        if path:
            return isdir(join(path, PROVENANCE_DIRNAME))
        return isdir(self.provenance_path)

    def connect_existing(self, path=None):
        if path:
            self.path = path
        if not self.has_provenance():
            print_msg('there is no provenance store in the current directory',
                True)
            sys.exit(1)
        self.connect()

    def query(self, text):
        with self.db_conn as db:
            for row in db.execute(text):
                yield row_to_dict(row)


# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import ast
import sys
import dis
import types
import itertools
import tokenize

from collections import OrderedDict

from ..cross_version import cross_compile, StringIO


class FunctionCall(ast.NodeVisitor):

    def __init__(self, visitor_class):
        self.func = []
        self.args = []
        self.keywords = {}
        self.starargs = []
        self.kwargs = []
        self.result = None
        self.visitor_class = visitor_class

    def all_args(self):
        return list(itertools.chain(
            itertools.chain.from_iterable(self.args),
            self.starargs,
            self.kwargs,
            itertools.chain.from_iterable(self.keywords.values())
        ))

    def use_visitor(self, node):
        visitor = self.visitor_class()
        visitor.visit(node)
        return [x if isinstance(x, FunctionCall) else x[0]
                for x in visitor.names]

    def visit_Call(self, node):
        self.func = self.use_visitor(node.func)
        self.args = [self.use_visitor(arg) for arg in node.args]
        for keyword in node.keywords:
            self.visit(keyword)
        if node.starargs:
            self.starargs = self.use_visitor(node.starargs)
        if node.kwargs:
            self.kwargs = self.use_visitor(node.kwargs)

    def visit_keyword(self, node):
        self.keywords[node.arg] = self.use_visitor(node.value)

    def info(self):
        result = ("line={}, col={}, "
                  "func={}, args={}, keywords={}, *args={}, **kwargs={}")
        return result.format(self.line, self.col, self.func, self.args,
                             self.keywords, self.starargs, self.kwargs)

    def __repr__(self):
        return "F({})".format(self.info())


class ClassDef(FunctionCall):

    def __repr__(self):
        return "Class()"


class Decorator(FunctionCall):

    def __init__(self, *args, **kwargs):
        super(Decorator, self).__init__(*args, **kwargs)
        self.fn = True

    def __repr__(self):
        return "Decorator({})".format(self.info())

    def visit_Name(self, node):
        self.func = self.use_visitor(node)
        self.fn = False

    def info(self):
        if self.fn:
            return super(Decorator, self).info()
        return "line={}, col={}, name={}".format(
            self.line, self.col, self.func)


class Generator(FunctionCall):

    def __init__(self, *args, **kwargs):
        self.type = args[-1]
        args = args[:-1]
        super(Generator, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "Generator({})".format(self.info())

    def info(self):
        return "line={}, col={}, type={}".format(
            self.line, self.col, self.type)


class Assert(FunctionCall):

    def __init__(self, *args, **kwargs):
        self.msg = args[-1]
        args = args[:-1]
        super(Assert, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "Assert({})".format(self.info())

    def info(self):
        return "line={}, col={}, msg={}".format(
            self.line, self.col, self.msg)


def index(lis, alternatives):
    for alt in alternatives:
        try:
            return lis.index(alt)
        except ValueError:
            pass
    return None


def get_code_object(obj, compilation_mode="exec"):
    if isinstance(obj, types.CodeType):
        return obj
    elif isinstance(obj, types.FrameType):
        return obj.f_code
    elif isinstance(obj, types.FunctionType):
        return obj.__code__
    elif isinstance(obj, str):
        try:
            return cross_compile(obj, "<string>", compilation_mode)
        except SyntaxError as error:
            raise ValueError("syntax error in passed string")
    else:
        raise TypeError("get_code_object() can not handle '%s' objects" %
                        (type(obj).__name__,))


def diss(obj, mode="exec", recurse=False):
    _visit(obj, dis.dis, mode, recurse)


def ssc(obj, mode="exec", recurse=False):
    _visit(obj, dis.show_code, mode, recurse)


def _visit(obj, visitor, mode="exec", recurse=False):
    obj = get_code_object(obj, mode)
    visitor(obj)
    if recurse:
        for constant in obj.co_consts:
            if type(constant) is type(obj):
                _visit(constant, visitor, mode, recurse)

def safeget(container, index):
    try:
        return container[index]
    except IndexError as err:
        if not err.args:
            err.args = ('',)
        err.args = (err.args[0] + '\n Get\n  Index {}\n  Container {}'.format(
            index, container),) + err.args[1:]
        raise err



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from ..persistence import row_to_dict, persistence

class Activation(dict):

    def el(self, e):
        if isinstance(e, dict):
            return Activation(e)
        else:
            return e

    def __key(self):
        return tuple((k,self.el(self[k])) for k in sorted(self))
    def __hash__(self):
        return hash(self.__key())
    def __eq__(self, other):
        return self.__key() == other.__key()

    def objects(self):
        return map(row_to_dict, persistence.load(
            'object_value', function_activation_id=self['id']))



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import sys
import argparse

from ...cmd import Run, LAST_TRIAL
from ...persistence import persistence
from ...models.trial import Trial


from .command import IpythonCommandMagic


class NowRun(IpythonCommandMagic, Run):
    """Run cells with noworkflow in a subprocess

    This uses the %%script magic.
    Please look at %%script? to understand --proc, --bg, --out and --err.

    Examples
    --------
    ::
        In [1]: my_var = 2

        In [2]: %%now_run --name script1 $my_var 30
           ...: import sys
           ...: print(sys.argv[1])
           ...: print(sys.argv[2])
        2
        30
        Out[2]: <Trial 1>

    ::
        In [3]: %%now_run --name script2 -v
           ...: l = range(4)
           ...: c = sum(l)
           ...: print(c)
        [now] removing noWorkflow boilerplate
        [now] setting up local provenance store
        [now] collecting definition provenance
        [now]   registering user-defined functions
        [now] collecting deployment provenance
        [now]   registering environment attributes
        [now]   searching for module dependencies
        [now]   registering provenance from 0 modules
        [now] collection execution provenance
        [now]   executing the script
        6
        [now] the execution of trial 2 finished successfully
        Out[3]: <Trial 2>

        In [4]: trial = _

    ::
        In [5]: %now_run script1.py $my_var 30
        Out[5]: <Trial 1>

    ::
        In [6]: trial = %now_run --name script2 script2.py
        6

    """

    def add_arguments(self):
        super(NowRun, self).add_arguments()
        add_arg = self.add_argument
        add_arg('--out', type=str,
                help="""The variable in which to store stdout from the script.
                If the script is backgrounded, this will be the stdout *pipe*,
                instead of the stdout text itself.
                """)
        add_arg('--err', type=str,
                help="""The variable in which to store stderr from the script.
                If the script is backgrounded, this will be the stderr *pipe*,
                instead of the stderr text itself.
                """)
        add_arg('--bg', action="store_true",
                help="""Whether to run the script in the background.
                If given, the only way to see the output of the command is
                with --out/err.
                """)
        add_arg('--proc', type=str,
                help="""The variable in which to store Popen instance.
                This is used only when --bg option is given.
                """)
        add_arg('params', nargs=argparse.REMAINDER,
                help='params to be passed to script')

    def execute(self, func, line, cell, magic_cls):
        # Calculate noworkflow params and script params
        argv, args = self.arguments(func, line)
        params = args.params
        if params:
            argv = argv[:-len(params)]
        # Create tmp file
        directory = args.dir or os.path.abspath(os.path.curdir)
        original_path = persistence.path
        try:
            persistence.connect(directory)
            filename = ''
            if cell:
                #cell = cell.encode('utf8', 'replace')
                filename = magic_cls.shell.mktempfile(data=cell, prefix='now_run_')

            # Set execution line
            cmd = "now run --create_last {directory} {args} {script} {params}".format(
                directory='' if args.dir else '--dir {0}'.format(directory),
                args=' '.join(argv),
                script=filename,
                params=' '.join(params)
            )
            script = magic_cls.shell.find_cell_magic('script').__self__
            result = script.shebang(cmd, "")
            tmp_dir = os.path.dirname(filename or params[0])

            try:
                with open(os.path.join(tmp_dir, LAST_TRIAL), 'r') as f:
                    return Trial(trial_id=int(f.read()))
            except e:
                print('Failed', e)
        finally:
            persistence.connect(original_path)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


from ..models import Trial, Diff, History, Model, TrialProlog


models = {
    'History': [History],
    'Trial': [Trial],
    'Diff': [Diff],
    '*': [History, Trial, Diff],
}


def update_all(attribute, value, model='*'):
    for cls in models[model]:
        for instance in cls.get_instances():
            setattr(instance, attribute, value)


def set_default(attribute, value, all=False, model='*'):
    for cls in models[model]:
        if attribute in cls.DEFAULT:
            cls.DEFAULT[attribute] = value

    if all:
        update_all(attribute, value, model=model)



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import time

from collections import defaultdict, OrderedDict
from .model import Model
from .utils import calculate_duration, FORMAT
from ..persistence import row_to_dict
from ..persistence import persistence as pers


class History(Model):
    """ This model represents the workflow evolution history

    It is possible to filter the evolution history by selecting the script:
        history.script = "script1.py"

    The list of scripts can be accessed by:
        history.scripts()

    It is also possible to filter the evolution history by selecting the
    trial status:
        history.execution = "finished"

    The list of status are:
        finished: show only finished trials
        unfinished: show only unfinished trials
        backup: show only backup trials

    The default option for both filters is "*", which means that all trials
    appear in the history
        history.script = "*"
        history.execution = "*"

    You can change the graph width and height by the variables:
        history.graph_width = 600
        history.graph_height = 200
    """

    DEFAULT = {
        'graph_width': 700,
        'graph_height': 300,
        'script': '*',
        'execution': '*',
        'data': {},
    }

    def __init__(self, **kwargs):
        super(History, self).__init__(**kwargs)
        self.initialize_default(kwargs)

        self.execution_options = ["*", "finished", "unfinished", "backup"]

    def scripts(self):
        """ Returns the list of scripts used for trials """
        return {s[0].rsplit('/', 1)[-1] for s in pers.distinct_scripts()}

    def graph_data(self, script="*", execution="*"):
        """ Prepares evolution history as a dict """
        if self.script != "*" and script == "*":
            script = self.script
        if self.execution != "*" and execution == "*":
            execution = self.execution

        key = (script, execution)
        if key in self.data:
            return self.data[key]

        nodes, edges = [], []
        result = {'nodes': nodes, 'edges': edges}
        id_map, children = {}, defaultdict(list)
        scripts, order = defaultdict(list), OrderedDict()

        # Filter nodes and adds to dicts
        tid = 0
        for trial in map(row_to_dict, pers.load('trial', order="start")):
            different_script = (trial['script'] != script)
            finished = trial['finish']
            unfinished = not finished and trial['run']
            backup = not finished and not trial['run']

            if script != '*' and different_script:
                continue
            if execution == 'finished' and not finished:
                continue
            if execution == 'unfinished' and not unfinished:
                continue
            if execution == 'backup' and not backup:
                continue

            trial_id = trial["id"]
            trial["level"] = 0
            trial["status"] = "Finished" if trial["finish"] else "Unfinished"
            if not trial['run']:
                trial["status"] = "Backup"
            trial["tooltip"] = """
                <b>{script}</b><br>
                {status}<br>
                Start: {start}<br>
                Finish: {finish}
                """.format(**trial)
            if trial['finish']:
                trial["tooltip"] += """
                <br>
                Duration: {duration}ns
                """.format(duration=calculate_duration(trial))
            id_map[trial_id] = tid
            scripts[trial['script']].append(trial)
            nodes.append(trial)
            tid += 1

        # Create edges
        for trial in reversed(nodes):
            trial_id, parent_id = trial["id"], trial["parent_id"]
            if parent_id and parent_id in id_map:
                edges.append({
                    'source': id_map[trial_id],
                    'target': id_map[parent_id],
                    'right': 1,
                    'level': 0
                })
                children[parent_id].append(trial_id)
            order[trial['script']] = 1

        # Set position
        level = 0
        for script in order:
            last = level
            for trial in scripts[script]:
                trial_id, parent_id = trial["id"], trial["parent_id"]
                if parent_id and parent_id in id_map:
                    parent = nodes[id_map[parent_id]]
                    if children[parent_id].index(trial_id) > 0:
                        trial["level"] = last
                        last += 1
                    else:
                        trial["level"] = parent["level"]
                    level = max(level, trial["level"] + 1)
                else:
                    trial["level"] = level
                    level += 1
                    last += 1

        self.data[key] = result
        return result

    def _repr_html_(self):
        """ Displays d3 graph on ipython notebook """
        uid = str(int(time.time()*1000000))

        result = """
            <div class="nowip-history" data-width="{width}"
                 data-height="{height}" data-uid="{uid}">
                {data}
            </div>
        """.format(
            uid=uid,
            data=self.escape_json(self.graph_data()),
            width=self.graph_width, height=self.graph_height)
        return result



# Copyright (c) 2015 Universidade Federal Fluminense (UFF)
# Copyright (c) 2015 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse

from IPython.core.display import display_javascript
from IPython.utils.text import DollarFormatter

from ...persistence import persistence
from ...formatter import Table
from ..models import set_default
from .command import IpythonCommandMagic


class NowSQL(IpythonCommandMagic):
    """Queries the provenance database with SQL

    Examples
    --------
    ::
        In  [1]: %%now_sql
            ...: SELECT DISTINCT script FROM trial
        Out [1]: [['script'],
            ...:  ['script1.py'],
            ...:  ['script2.py']]

        In  [2]: %%now_sql --result tupleit
            ...: SELECT * FROM trial

        In  [3]: tupleit
        Out [3]: <generator object query at 0x7fd9ebdbebe0>
    """

    def __init__(self, *args, **kwargs):
        super(NowSQL, self).__init__(*args, **kwargs)
        js = """
            var mode = 'magic_text/x-sql';
            if (!IPython.CodeCell.config_defaults.highlight_modes[mode]) {
                IPython.CodeCell.config_defaults.highlight_modes[mode] = {
                    'reg':[]
                };
            }
            IPython.CodeCell.config_defaults.highlight_modes[mode].reg.push(
                /^%%now_sql/
            );
        """;
        display_javascript(js, raw=True)


    def add_arguments(self):
        super(NowSQL, self).add_arguments()
        add_arg = self.add_argument
        add_arg('--result', type=str,
                help="""The variable in which the result will be stored""")

    def execute(self, func, line, cell, magic_cls):
        f = DollarFormatter()
        cell = f.vformat(cell, args=[], kwargs=magic_cls.shell.user_ns.copy())
        _, args = self.arguments(func, line)
        result = persistence.query(cell)
        if args.result:
            magic_cls.shell.user_ns[args.result] = result
        else:
            result = list(result)
            table = Table()
            if result:
                table.append(result[0].keys())
            for line in result:
                table.append(line.values())
            return table



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


class Activation(object):
    __slots__ = (
        'start', 'file_accesses', 'function_activations',
        'finish', 'return_value',
        'name', 'line', 'arguments', 'globals',
        'context', 'slice_stack', 'lasti',
        'args', 'kwargs', 'starargs',
    )

    def __init__(self, name, line, lasti):
        self.start = 0.0
        self.file_accesses = []
        self.function_activations = []
        self.finish = 0.0
        self.return_value = None
        self.name = name
        self.line = line
        self.arguments = {}
        self.globals = {}
        # Variable context. Used in the slicing lookup
        self.context = {}
        # Line execution stack.
        # Used to evaluate function calls before execution line
        self.slice_stack = []
        self.lasti = lasti

        self.args = []
        self.kwargs = []
        self.starargs = []



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


class Command(object):

    def __init__(self, help, cmd=None):
        self.cmd = cmd or type(self).__name__.lower()
        self.help = help

    def create_parser(self, subparsers):
        kwargs = {}
        if self.help:
            kwargs['help'] = self.help
        self.parser = subparsers.add_parser(self.cmd, **kwargs)

        self.add_arguments()
        self.parser.set_defaults(func=self.execute)

    def add_arguments(self):
        pass

    def add_argument(self, *args, **kwargs):
        return self.parser.add_argument(*args, **kwargs)

    def add_argument_cmd(self, *args, **kwargs):
        return self.parser.add_argument(*args, **kwargs)

    def execute(self, *args, **kwargs):
        abstract()



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from datetime import datetime
from collections import namedtuple, defaultdict

from ..persistence import persistence
from .utils import calculate_duration, FORMAT


Edge = namedtuple("Edge", "node count")


class TrialGraphVisitor(object):

    def __init__(self):
        self.nodes = []
        self.edges = []
        self.delegated = {
            'initial': Edge(0, 1)
        }
        self.nid = 0
        self.min_duration = defaultdict(lambda: 1000^10)
        self.max_duration = defaultdict(lambda: 0)
        self.keep = None

    def update_durations(self, duration, tid):
        self.max_duration[tid] = max(self.max_duration[tid], duration)
        self.min_duration[tid] = min(self.min_duration[tid], duration)

    def update_node(self, node):
        node['mean'] = node['duration'] / node['count']
        node['info'].update_by_node(node)
        node['info'] = repr(node['info'])

    def to_dict(self):
        for node in self.nodes:
            n = node['node']
            self.update_node(n)
            self.update_durations(n['duration'], n['trial_id'])

        self.update_edges()
        return {
            'nodes': self.nodes,
            'edges': self.edges,
            'min_duration': self.min_duration,
            'max_duration': self.max_duration
        }

    def update_edges(self):
        for edge in self.edges:
            if edge['type'] in ['return', 'call']:
                edge['count'] = ''

    def add_node(self, node):
        self.nodes.append(node.to_dict(self.nid))
        original = self.nid
        self.nid += 1
        return original


    def add_edge(self, source, target, count, typ):
        self.edges.append({
            'source': source,
            'target': target,
            'count': count,
            'type': typ
        })

    def use_delegated(self):
        result = self.delegated
        self.delegated = {}
        return result

    def solve_delegation(self, node_id, node_count, delegated):
        self.solve_cis_delegation(node_id, node_count, delegated)
        self.solve_ret_delegation(node_id, node_count, delegated)

    def solve_cis_delegation(self, node_id, node_count, delegated):
        # call initial sequence
        for typ in ['call', 'initial', 'sequence']:
            if typ in delegated:
                edge = delegated[typ]
                self.add_edge(edge.node, node_id, node_count, typ)

    def solve_ret_delegation(self, node_id, node_count, delegated):
        if 'return' in delegated:
            edge = delegated['return']
            self.add_edge(node_id, edge.node, edge.count, 'return')

    def visit_call(self, call):
        delegated = self.use_delegated()
        caller_id = self.add_node(call.caller)
        self.nodes[caller_id]['repr'] = repr(call)

        if delegated:
            self.solve_delegation(caller_id, 0, delegated)

        self.delegated['call'] = Edge(caller_id, 1)
        self.delegated['return'] = Edge(caller_id, 1)

        call.called.visit(self)
        return caller_id, call

    def visit_group(self, group):
        delegated = self.use_delegated()

        node_map = {}
        for element in group.nodes.values():
            node_id, node = element.visit(self)
            node_map[node] = node_id

        self.solve_cis_delegation(node_map[group.next], group.count, delegated)
        self.solve_ret_delegation(node_map[group.last], group.count, delegated)

        for previous, edges in group.edges.items():
            for next, count in edges.items():
                self.add_edge(node_map[previous], node_map[next],
                              count, 'sequence')

        return node_map[group.next], group.next

    def visit_single(self, single):
        delegated = self.use_delegated()
        node_id = self.add_node(single)
        self.nodes[node_id]['repr'] = repr(single)

        if delegated:
            self.solve_delegation(node_id, single.count, delegated)
        return node_id, single

    def visit_mixed(self, mixed):
        mixed.mix_results()
        node_id, node = mixed.first.visit(self)
        self.nodes[node_id]['duration'] = mixed.duration
        return node_id, node

    def visit_default(self, empty):
        return None, None


class TrialGraphCombineVisitor(TrialGraphVisitor):

    def __init__(self):
        super(TrialGraphCombineVisitor, self).__init__()
        self.context = {}
        self.context_edges = {}
        self.namestack = []

    def update_edges(self):
        pass

    def namespace(self):
        return ' '.join(self.namestack)

    def update_namespace_node(self, node, single):
        node['count'] += single.count
        node['duration'] += single.duration
        node['info'].add_activation(single.activation)

    def add_node(self, single):
        self.namestack.append(single.name_id())
        namespace = self.namespace()
        self.namestack.pop()
        if namespace in self.context:
            context = self.context[namespace]
            self.update_namespace_node(context['node'], single)

            return self.context[namespace]['index']

        single.namespace = namespace
        result = super(TrialGraphCombineVisitor, self).add_node(single)
        self.context[namespace] = self.nodes[-1]
        return result

    def add_edge(self, source, target, count, typ):

        edge = "{} {} {}".format(source, target, typ)

        if not edge in self.context_edges:
            super(TrialGraphCombineVisitor, self).add_edge(source, target,
                                                           count, typ)
            self.context_edges[edge] = self.edges[-1]
        else:
            e = self.context_edges[edge]
            self.context_edges[edge]['count'] += count

    def visit_call(self, call):
        self.namestack.append(call.caller.name_id())
        result = super(TrialGraphCombineVisitor, self).visit_call(call)
        self.namestack.pop()
        return result

    def visit_mixed(self, mixed):
        node_id, node = None, None
        for element in mixed.elements:
            node_id, node = element.visit(self)
        return node_id, node



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import linecache
import itertools
import numbers
from collections import namedtuple
from datetime import datetime
from .profiler import Profiler
from ..utils import print_fn_msg
from ..prov_definition import SlicingVisitor
from ..prov_definition import (FunctionCall, ClassDef, Decorator,
                               Generator, Assert)

from ..persistence import persistence

try:
    immutable = (bool, numbers.Number, str, unicode)
except NameError:
    immutable = (bool, numbers.Number, str, bytes)

WITHOUT_PARAMS = (ClassDef, Assert, Generator)

class Variable(object):
    __slots__ = (
        'id',
        'name',
        'line',
        'value',
        'time',
    )

    def __init__(self, id, name, line, value=None, time=None):
        self.id = id
        self.name = name
        self.line = line
        self.value = value
        self.time = time

    def __repr__(self):
        return "Variable(id={}, name={}, line={}, value={})".format(
            self.id, self.name, self.line, self.value)

#Variable = namedtuple("Variable", "id name line value time")
Dependency = namedtuple("Dependency", "id dependent supplier")
Usage = namedtuple("Usage", "id vid name line")
Return = namedtuple("Return", "activation var")

class ObjectStore(object):

    def __init__(self, cls):
        self.cls = cls
        self.store = []
        self.id = -1

    def __getitem__(self, index):
        return self.store[index]

    def add(self, *args):
        self.id += 1
        self.store.append(self.cls(self.id, *args))
        return self.id

    def remove(self, value):
        self.store.remove(value)

    def __iter__(self):
        return self.store.__iter__()


class Tracer(Profiler):

    def __init__(self, *args):
        super(Tracer, self).__init__(*args)
        definition = self.metascript['definition']
        self.event_map['line'] = self.trace_line

        # Store slicing provenance
        self.variables = ObjectStore(Variable)
        self.dependencies = ObjectStore(Dependency)
        self.usages = ObjectStore(Usage)

        # Returns
        self.returns = {}
        # Avoid using the same event for tracer and profiler
        self.last_event = None

        # Useful maps
        # Map of dependencies by line
        self.line_dependencies = definition.line_dependencies[self.script]
        # Map of name_refs by line
        self.line_usages = definition.line_usages[self.script]
        # Map of calls by line and col
        self.call_by_col = definition.call_by_col[self.script]
        # Map of calls by line and lasti
        self.call_by_lasti = definition.call_by_lasti[self.script]
        # Set of imports
        self.imports = definition.imports[self.script]

    def add_variable(self, name, line, f_locals, value='--check--'):
        """ Adds variable """
        if value == '--check--' and name in f_locals:
            value = repr(f_locals[name])
        else:
            value = 'now(n/a)'
        return self.variables.add(name, line, value, datetime.now())

    def add_dependency(self, var, dep, activation):
        """ Adds dependency """
        var_id, dependencies_add = var.id, self.dependencies.add
        context = activation.context

        # Variable
        if dep in context:
            dependencies_add(var_id, context[dep].id)
        # Function Call
        if isinstance(dep, tuple):
            call = self.call_by_col[dep[0]][dep[1]]
            lasti, returns = call.lasti, self.returns

            if not lasti in returns:
                return
            _return = returns[lasti]

            vid = self.add_variable('call {}'.format(_return.activation.name),
                                    dep[0], {}, 'now(n/a)')
            # Call was not evaluated before
            #if call.result is None or _return.var.line == call.result[1]:
            del returns[lasti]
            call.result = (_return.var.id, _return.var.line)
            dependencies_add(vid, _return.var.id)

            self.add_dependencies(self.variables[vid], activation, call.func)
            dependencies_add(var_id, vid)

    def add_dependencies(self, var, activation, dependencies):
        """ Adds dependencies to var """
        add_dependency = self.add_dependency
        for dep in dependencies:
            add_dependency(var, dep, activation)

    def slice_line(self, activation, lineno, f_locals, filename):
        """ Generates dependencies from line """
        print_fn_msg(lambda: 'Slice [{}] -> {}'.format(lineno,
                linecache.getline(filename, lineno).strip()))

        context, variables = activation.context, self.variables
        usages_add = self.usages.add
        add_variable = self.add_variable
        add_dependencies = self.add_dependencies

        dependencies = self.line_dependencies[lineno]
        usages = self.line_usages[lineno]['Load']

        for name in usages:
            if name in context:
                usages_add(context[name].id, name, lineno)

        for name, others in dependencies.items():
            if name == 'return':
                vid = add_variable(name, lineno, f_locals,
                                   value=activation.return_value)
                self.returns[activation.lasti] = Return(activation,
                                                        variables[vid])
            else:
                vid = add_variable(name, lineno, f_locals)
            add_dependencies(variables[vid], activation, others)

            context[name] = variables[vid]

    def close_activation(self, event, arg):
        """ Slice all lines from closing activation """
        for line in self.activation_stack[-1].slice_stack:
            self.slice_line(*line)
        super(Tracer, self).close_activation(event, arg)

    def add_generic_return(self, frame, event, arg, ccall=False):
        """ Add return to functions that do not have return
            For ccall, add dependency from all params
        """

        lineno = frame.f_lineno
        variables = self.variables

        # Artificial return condition
        activation = self.activation_stack[-1]
        lasti = activation.lasti
        if not 'return' in self.line_dependencies[lineno]:
            vid = self.add_variable('return', lineno, {},
                                    value=activation.return_value)
            activation.context['return'] = variables[vid]
            self.returns[lasti] = Return(activation, variables[vid])

            if ccall and frame.f_code.co_filename == self.script:
                caller = self.activation_stack[-2]
                call = self.call_by_lasti[lineno][lasti]
                all_args = call.all_args()
                self.add_dependencies(variables[vid], caller, all_args)
                self.add_inter_dependencies(frame, all_args, caller)

    def match_arg(self, passed, arg, caller, activation, line, f_locals):
        """ Matches passed param with argument """
        context = activation.context

        if arg in context:
            act_var = context[arg]
        else:
            vid = self.add_variable(arg, line, f_locals)
            act_var = self.variables[vid]
            context[arg] = act_var

        if passed:
            self.add_dependency(act_var, passed, caller)

    def match_args(self, args, param, caller, activation, line, f_locals):
        """ Matches passed param with arguments """
        for arg in args:
            self.match_arg(arg, param, caller, activation, line, f_locals)

    def add_inter_dependencies(self, frame, args, activation):
        """ Adds dependencies between params """
        global immutable
        f_locals, context = frame.f_locals, activation.context
        lineno, variables = frame.f_lineno, self.variables
        add_variable = self.add_variable
        add_dependencies = self.add_dependencies

        added = {}
        for arg in args:
            try:
                var = f_locals[arg]
                if not isinstance(var, immutable):
                    vid = add_variable(arg, lineno, {}, value=var)
                    variable = variables[vid]
                    add_dependencies(variable, activation, args)
                    added[arg] = variable
            except KeyError:
                pass

        for arg, variable in added.items():
            context[arg] = variable

    def add_argument_variables(self, frame):
        """ Adds argument variables """
        back = frame.f_back
        call = self.call_by_lasti[back.f_lineno][back.f_lasti]
        if isinstance(call, WITHOUT_PARAMS):
            return
        if isinstance(call, Decorator) and not call.fn:
            return
        f_locals = frame.f_locals
        match_args = self.match_args
        match_arg = self.match_arg
        caller, act = self.activation_stack[-2:]
        act_args_index = act.args.index
        line = frame.f_lineno
        sub = -[bool(act.starargs), bool(act.kwargs)].count(True)

        order = act.args + act.starargs + act.kwargs
        used = [0 for _ in order]
        j = 0
        for i, call_arg in enumerate(call.args):
            j = i if i < len(order) + sub else sub
            act_arg = order[j]
            match_args(call_arg, act_arg, caller, act, line, f_locals)
            used[j] += 1
        for act_arg, call_arg in call.keywords.items():
            try:
                i = act_args_index(act_arg)
                match_args(call_arg, act_arg, caller, act, line, f_locals)
                used[i] += 1
            except ValueError:
                for kw in act.kwargs:
                    match_args(call_arg, kw, caller, act, line, f_locals)

        # ToDo: improve matching
        #   Ignore default params
        #   Do not match f(**kwargs) with def(*args)
        args = [(i, order[i]) for i in range(len(used)) if not used[i]]
        for star in call.kwargs + call.starargs:
            for i, act_arg in args:
                match_args(star, act_arg, caller, act, line, f_locals)
                used[i] += 1

        args = [(i, order[i]) for i in range(len(used)) if not used[i]]
        for i, act_arg in args:
            match_arg(None, act_arg, caller, act, line, f_locals)

        self.add_inter_dependencies(back, call.all_args(), caller)

    def trace_call(self, frame, event, arg):
        super(Tracer, self).trace_call(frame, event, arg)
        if self.script != frame.f_back.f_code.co_filename:
            return
        if frame.f_back.f_lineno in self.imports:
            return
        self.add_argument_variables(frame)

    def trace_c_return(self, frame, event, arg):
        activation = self.activation_stack[-1]
        self.add_generic_return(frame, event, arg, ccall=True)
        super(Tracer, self).trace_c_return(frame, event, arg)
        activation.context['return'].value = activation.return_value

    def trace_return(self, frame, event, arg):
        activation = self.activation_stack[-1]
        self.add_generic_return(frame, event, arg)
        super(Tracer, self).trace_return(frame, event, arg)
        activation.context['return'].value = activation.return_value

    def trace_line(self, frame, event, arg):
        # Different file
        if frame.f_code.co_filename != self.script:
            return

        lineno = frame.f_lineno

        activation = self.activation_stack[-1]

        print_fn_msg(lambda: '[{}] -> {}'.format(lineno,
                linecache.getline(self.script, lineno).strip()))

        if activation.slice_stack:
            self.slice_line(*activation.slice_stack.pop())
        activation.slice_stack.append([
            activation, lineno, frame.f_locals, self.script])

    def tracer(self, frame, event, arg):
        current_event = (event, frame.f_lineno, frame.f_code)
        if self.last_event != current_event:
            self.last_event = current_event
            return super(Tracer, self).tracer(frame, event, arg)
        return self.tracer

    def tearup(self):
        sys.settrace(self.tracer)
        sys.setprofile(self.tracer)

    def store(self):
        while self.activation_stack:
            self.close_activation('store', None)
        super(Tracer, self).store()
        persistence.store_slicing(self.trial_id, self.variables,
                                  self.dependencies, self.usages)
        for var in self.variables:
            print_fn_msg(lambda: var)

        for dep in self.dependencies:
            print_fn_msg(lambda: "{}\t<-\t{}".format(
                                        self.variables[dep.dependent],
                                        self.variables[dep.supplier]))

        for var in self.usages:
            print_fn_msg(lambda: var)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref
import json
from collections import defaultdict


class Model(object):

    __refs__ = defaultdict(list)
    def __init__(self, *args, **kwargs):
        self.__refs__[self.__class__].append(weakref.ref(self))

    def initialize_default(self, kwargs):
        for key, value in self.DEFAULT.items():
            setattr(self, key, value)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def escape_json(self, data):
        data = json.dumps(data)
        return (data.replace('&', '\\u0026')
                    .replace('<', '\\u003c')
                    .replace('>', '\\u003e'))

    @classmethod
    def get_instances(cls):
        for inst_ref in cls.__refs__[cls]:
            inst = inst_ref()
            if inst is not None:
                yield inst

    @classmethod
    def all_models(cls):
        for c in cls.__refs__:
            for instance in c.get_instances():
                yield instance



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import pyposast
from datetime import datetime

from ..utils import print_msg
from ..persistence import persistence
from .function_visitor import FunctionVisitor
from .slicing_visitor import SlicingVisitor
from .definition import Definition
from .utils import FunctionCall, ClassDef, Decorator, Generator, Assert



def visit_ast(metascript):
    '''returns a visitor that visited the tree and filled the attributes:
        functions: map of function in the form:
            name -> (arguments, global_vars, calls, code_hash)
        name_refs[path]: map of identifiers in categories Load, Store
        dependencies[path]: map of dependencies
    '''
    tree = pyposast.parse(metascript['code'], metascript['path'])
    visitor = SlicingVisitor(metascript)
    visitor.result = visitor.visit(tree)
    visitor.extract_disasm()
    visitor.teardown()
    return visitor


def collect_provenance(args, metascript):
    now = datetime.now()
    try:
        metascript['trial_id'] = persistence.store_trial(
            now, metascript['name'], metascript['code'], ' '.join(sys.argv[1:]),
            args.bypass_modules)
    except TypeError as e:
        if args.bypass_modules:
            print_msg('not able to bypass modules check because no previous '
                      'trial was found', True)
            print_msg('aborting execution', True)
        else:
            raise e

        sys.exit(1)
    definition = Definition(metascript)
    print_msg('  registering user-defined functions')
    visitor = visit_ast(metascript)
    persistence.store_function_defs(metascript['trial_id'], visitor.functions)
    if args.disasm:
        print('\n'.join(visitor.disasm))
    definition.add_visitor(visitor)
    metascript['definition'] = definition



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import sys

from .. import utils
from ..persistence import persistence
from ..models.trial import Trial
from .command import Command

class Show(Command):

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg('trial', type=int, nargs='?',
                help='trial id or none for last trial')
        add_arg('-m', '--modules', action='store_true',
                help='shows module dependencies')
        add_arg('-d', '--function-defs', action='store_true',
                help='shows the user-defined functions')
        add_arg('-e', '--environment', action='store_true',
                help='shows the environment conditions')
        add_arg('-a', '--function-activations', action='store_true',
                help='shows function activations')
        add_arg('-f', '--file-accesses', action='store_true',
                help='shows read/write access to files')
        add_arg('--dir', type=str,
                help='set project path where is the database. Default to '
                     'current directory')

    def wrap(self, *args, **kwargs):
        return utils.wrap(*args, **kwargs)

    def execute(self, args):
        persistence.connect_existing(args.dir or os.getcwd())
        trial = Trial(trial_id=args.trial, exit=True)
        self.print_trial(trial)

        if args.modules:
            self.print_modules(trial)

        if args.function_defs:
            self.print_function_defs(trial)

        if args.environment:
            self.print_environment(trial)

        if args.function_activations:
            self.print_function_activations(trial)

        if args.file_accesses:
            self.print_file_accesses(trial)

    def print_trial(self, trial):
        utils.print_msg('trial information:', True)
        print(self.wrap("""\
            Id: {id}
            Inherited Id: {inherited_id}
            Script: {script}
            Code hash: {code_hash}
            Start: {start}
            Finish: {finish}\
            """.format(**trial.info())))

    def print_modules(self, trial):
        a, b = trial.modules()
        modules = a + list(b)
        utils.print_msg('this trial depends on the following modules:', True)
        output = []
        for module in modules:
            output.append(self.wrap("""\
                Name: {name}
                Version: {version}
                Path: {path}
                Code hash: {code_hash}\
                """.format(**module), other="\n    "))
        print('\n\n'.join(output))

    def print_function_defs(self, trial):
        utils.print_msg('this trial has the following functions:', True)
        output = []
        for function_def in trial.function_defs().values():
            objects = {'GLOBAL':[], 'ARGUMENT':[], 'FUNCTION_CALL':[]}
            for obj in persistence.load('object',
                                        function_def_id=function_def['id']):
                objects[obj['type']].append(obj['name'])
            output.append(self.wrap("""\
                Name: {name}
                Arguments: {arguments}
                Globals: {globals}
                Function calls: {calls}
                Code hash: {code_hash}\
                """.format(arguments=', '.join(objects['ARGUMENT']),
                           globals=', '.join(objects['GLOBAL']),
                           calls=', '.join(objects['FUNCTION_CALL']),
                           **function_def), other="\n    "))
        print('\n\n'.join(output))

    def print_environment(self, trial):
        utils.print_map(
            'this trial has been executed under the following environment '
            'conditions', trial.environment())

    def print_function_activation(self, trial, activation, level = 1):
        object_values = {'GLOBAL':[], 'ARGUMENT':[]}
        name = {'GLOBAL':'Globals', 'ARGUMENT':'Arguments'}
        for obj in activation.objects():
            object_values[obj['type']].append(
                '{} = {}'.format(obj['name'], obj['value']))
        text = self.wrap(
            '{line}: {name} ({start} - {finish})'.format(**activation),
            initial='  ' * level)
        indent = text.index(': ') + 2
        print(text)
        for typ, values in object_values.items():
            print(self.wrap(
                '{type}: {values}'.format(type=name[typ],
                                          values=', '.join(values)),
                initial=' ' * indent))
        if activation['return']:
            print(self.wrap(
                'Return value: {ret}'.format(ret=activation['return']),
                initial=' ' * indent))

        for inner_activation in trial.activations(caller_id=activation['id']):
            self.print_function_activation(trial, inner_activation, level + 1)

    def print_function_activations(self, trial):
        utils.print_msg('this trial has the following function activation '
                        'graph:', True)

        for inner_function_activation in trial.activations(caller_id=None):
            self.print_function_activation(trial, inner_function_activation)

    def print_file_accesses(self, trial):
        utils.print_msg('this trial accessed the following files:', True)
        output = []
        for file_access in trial.file_accesses():
            output.append(self.wrap("""\
                Name: {name}
                Mode: {mode}
                Buffering: {buffering}
                Content hash before: {content_hash_before}
                Content hash after: {content_hash_after}
                Timestamp: {timestamp}
                Function: {stack}\
                """.format(**file_access)))
        print('\n\n'.join(output))


# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import hashlib
import os

from .provider import Provider


class StorageProvider(Provider):

    def put(self, content):
        content_hash = hashlib.sha1(content).hexdigest()
        content_dirname = os.path.join(self.content_path, content_hash[:2])
        if not os.path.isdir(content_dirname):
            os.makedirs(content_dirname)
        content_filename = os.path.join(content_dirname, content_hash[2:])
        if not os.path.isfile(content_filename):
            with self.std_open(content_filename, "wb") as content_file:
                content_file.write(content)
        return content_hash

    def get(self, content_hash):
        content_filename = os.path.join(self.content_path,
                                        content_hash[:2],
                                        content_hash[2:])
        with self.std_open(content_filename, 'rb') as content_file:
            return content_file.read()
